"""Functions for writing results to a file on disk."""
from __future__ import annotations

from pathlib import Path
from typing import Any
from typing import TextIO


def write_as_dict(filename: Path | str, *args: Any, data_format: str = ">10.8e") -> None:
    """Write arguments to a file as a dictionary."""
    with Path(filename).open("w") as ff:
        for name, value in args:
            ff.write(f"{name:<14} : {value:{data_format}}\n")


def write_as_list(
    filename: Path | str, *args: Any, header_format: str = "<15", data_format: str = ">+10.8e"
) -> None:
    """Write the arguments to a file as a list."""
    with Path(filename).open("w") as ff:
        _write(ff, header_format, [item for item in [arg[0] for arg in args]])
        for value in zip(*[arg[1] for arg in args]):
            _write(ff, data_format, value)


def _write(ff: TextIO, write_format: str, items: Any) -> None:
    """Write items to a file with a specific format."""
    if isinstance(items[0], str):
        ff.write("# ")
    else:
        ff.write("  ")
    ff.write("".join("{1:{0}} ".format(write_format, item) for item in items) + "\n")



"""Classes representing a pressure patch on the free surface."""
from __future__ import annotations

import abc
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import fmin

from planingfsi import math_helpers
from planingfsi import trig
from planingfsi import writers
from planingfsi.config import Config
from planingfsi.dictionary import load_dict_from_file
from planingfsi.potentialflow import pressureelement as pe

if TYPE_CHECKING:
    from planingfsi.fe.substructure import Interpolator
    from planingfsi.potentialflow.solver import PotentialPlaningSolver


class PressurePatch(abc.ABC):
    """Abstract base class representing a patch of pressure elements on the free surface.

    Attributes:
        pressure_elements: List of pressure elements.
        is_kutta_unknown: True of trailing edge pressure unknown.
        interpolator: Object to get interpolated body position if a `PlaningSurface`
        parent: Parent solver this patch belongs to.

    """

    name: str

    def __init__(self, *, parent: PotentialPlaningSolver | None) -> None:
        self.parent = parent
        self.pressure_elements: list[pe.PressureElement] = []
        self._end_pts = np.zeros(2)
        self.is_kutta_unknown = False
        self._neighbor_up: PressurePatch | None = None
        self._neighbor_down: PressurePatch | None = None
        self.interpolator: Interpolator | None = None

        self.drag_total = np.nan
        self.drag_pressure = np.nan
        self.drag_friction = np.nan
        self.lift_total = np.nan
        self.lift_pressure = np.nan
        self.lift_friction = np.nan
        self.moment_total = np.nan

    @property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        if self.parent is None:
            raise ValueError(
                "The parent must be set to access its configuration. "
                "This can be done either with the 'parent=parent' keyword argument, by setting it explicitly, "
                "or by using the `.add_planing_surface()` method."
            )
        return self.parent.config

    @property
    def base_pt(self) -> float:
        """The x-location of the base point."""
        return self._end_pts[0]

    @base_pt.setter
    def base_pt(self, x: float) -> None:
        self._end_pts[0] = x

    @property
    def length(self) -> float:
        """Length of pressure patch, which is the distance between end points."""
        return self._end_pts[1] - self._end_pts[0]

    @length.setter
    def length(self, value: float) -> None:
        raise NotImplementedError

    @property
    def neighbor_up(self) -> PressurePatch | None:
        """PressurePatch instance upstream of this one.

        When setting, this patch is set as the other's downstream neighbor.

        """
        return self._neighbor_up

    @neighbor_up.setter
    def neighbor_up(self, obj: PressurePatch | None) -> None:
        self._neighbor_up = obj
        if self._neighbor_up is not None:
            self._neighbor_up._neighbor_down = self

    @property
    def neighbor_down(self) -> PressurePatch | None:
        """PressurePatch instance downstream of this one.

        When setting, this patch is set as the other's upstream neighbor.

        """
        return self._neighbor_down

    @neighbor_down.setter
    def neighbor_down(self, obj: PressurePatch | None) -> None:
        self._neighbor_down = obj
        if self._neighbor_down is not None:
            self._neighbor_down._neighbor_up = self

    @property
    @abc.abstractmethod
    def element_coords(self) -> np.ndarray:
        """Get the x-position of pressure elements."""
        return NotImplemented

    def _reset_element_coords(self) -> None:
        """Re-distribute pressure element positions along the patch."""
        x = self.element_coords
        for i, el in enumerate(self.pressure_elements):
            el.x_coord = x[i]
            if not el.is_source and self.interpolator is not None:
                el.z_coord = self.interpolator.get_body_height(x[i])

            if isinstance(el, pe.CompleteTriangularPressureElement):
                # TODO: Remove ignore
                el.width = np.array([x[i] - x[i - 1], x[i + 1] - x[i]])  # type: ignore
            elif isinstance(el, pe.ForwardHalfTriangularPressureElement):
                el.width = x[i + 1] - x[i]  # type: ignore
            elif isinstance(el, pe.AftHalfTriangularPressureElement):
                el.width = x[i] - x[i - 1]  # type: ignore
            elif isinstance(el, pe.AftSemiInfinitePressureBand):
                el.width = np.inf  # type: ignore
            else:
                raise ValueError("Invalid Element Type!")

    def get_free_surface_height(self, x: float) -> float:
        """Get free surface height at a position x due to the elements on this patch.

        Args:
            x: x-position at which to calculate free-surface height.

        """
        return sum(el.get_influence(x) for el in self.pressure_elements)

    def calculate_forces(self) -> None:
        """Calculate the force components for this pressure patch."""

    @property
    def drag_wave(self) -> float:
        """Calculate wave drag of patch."""
        f = self.get_free_surface_height
        x_init = -10 * self.config.flow.lam
        (x_trough,) = fmin(f, x_init, disp=False)
        (x_crest,) = fmin(lambda x: -f(x), x_init, disp=False)
        return (
            0.0625
            * self.config.flow.density
            * self.config.flow.gravity
            * (f(x_crest) - f(x_trough)) ** 2
        )

    @property
    def _force_file_save_path(self) -> Path:
        """A path to the force file."""
        assert self.parent is not None
        return self.parent.simulation.it_dir / f"forces_{self.name}.{self.config.io.data_format}"

    def write_forces(self) -> None:
        """Write forces to file."""
        writers.write_as_dict(
            self._force_file_save_path,
            ["Drag", self.drag_total],
            ["WaveDrag", self.drag_wave],
            ["PressDrag", self.drag_pressure],
            ["FricDrag", self.drag_friction],
            ["Lift", self.lift_total],
            ["PressLift", self.lift_pressure],
            ["FricLift", self.lift_friction],
            ["Moment", self.moment_total],
            ["BasePt", self.base_pt],
            ["Length", self.length],
        )

    def load_forces(self) -> None:
        """Load forces from file."""
        dict_ = load_dict_from_file(self._force_file_save_path)
        self.drag_total = dict_.get("Drag", 0.0)
        self.drag_pressure = dict_.get("PressDrag", 0.0)
        self.drag_friction = dict_.get("FricDrag", 0.0)
        self.lift_total = dict_.get("Lift", 0.0)
        self.lift_pressure = dict_.get("PressLift", 0.0)
        self.lift_friction = dict_.get("FricLift", 0.0)
        self.moment_total = dict_.get("Moment", 0.0)
        self.base_pt = dict_.get("BasePt", 0.0)
        self.length = dict_.get("Length", 0.0)


class PressureCushion(PressurePatch):
    """Pressure Cushion consisting solely of source elements.

    Args:
        dict_: Dictionary containing patch definitions.

    Attributes:
        name: Name of patch.
        cushion_pressure: The value of the pressure in the cushion. If "Pc" is provided, the value will be
            loaded from the simulation-global `config.body.Pc` variable.
        cushion_type: The type of pressure cushion, either "finite" or "infinite".
        smoothing_factor: Used for standalone finite pressure cushions without upstream and downstream surfaces.
        num_elements: The number of pressure elements used to represent the pressure cushion.
        upstream_planing_surface: The `PlaningSurface` upstream of the cushion, used to determine bounds.
        downstream_planing_surface: The `PlaningSurface` downstream of the cushion, used to determine bounds.
        upstream_loc: If there is no upstream `PlaningSurface` instance, this is the fixed upstream bound.
        downstream_loc: If there is no downstream `PlaningSurface` instance, this is the fixed downstream bound.
        parent: The `PotentialPlaningSolver` to which this `PressureCushion` belongs.

    """

    _count = 0

    def __init__(
        self,
        *,
        name: str = "",
        cushion_pressure: float | Literal["Pc"] | None = None,
        cushion_type: Literal["finite"] | Literal["infinite"] = "finite",
        smoothing_factor: float = np.nan,
        num_elements: int = 10,
        upstream_planing_surface: PlaningSurface | str | None = None,
        downstream_planing_surface: PlaningSurface | str | None = None,
        upstream_loc: float | None = None,
        downstream_loc: float | None = None,
        parent: PotentialPlaningSolver | None = None,
        **_: Any,
    ) -> None:
        super().__init__(parent=parent)
        PressureCushion._count += 1
        self.name = name or f"pressureCushion{PressureCushion._count}"

        if cushion_pressure == "Pc" or cushion_pressure is None:
            self.cushion_pressure = getattr(self.config.body, "Pc", 0.0)
        else:
            self.cushion_pressure = cushion_pressure

        if isinstance(upstream_planing_surface, PlaningSurface):
            self.neighbor_up = upstream_planing_surface
        else:
            self.neighbor_up = [
                surf
                for surf in self.parent.planing_surfaces
                if surf.name == upstream_planing_surface
            ][0]

        if isinstance(downstream_planing_surface, PlaningSurface):
            self.neighbor_down = downstream_planing_surface
        else:
            self.neighbor_down = [
                surf
                for surf in self.parent.planing_surfaces
                if surf.name == downstream_planing_surface
            ][0]

        if self.neighbor_down is not None:
            self.neighbor_down.upstream_pressure = self.cushion_pressure
        if self.neighbor_up is not None:
            self.neighbor_up.kutta_pressure = self.cushion_pressure

        # TODO: cushion_type variability should be managed by sub-classing PressurePatch
        self.cushion_type = cushion_type
        if self.cushion_type == "infinite":
            # Dummy element, will have 0 pressure
            self.pressure_elements.append(
                pe.AftSemiInfinitePressureBand(is_source=True, is_on_body=False, parent=self)
            )
            self.pressure_elements.append(
                pe.AftSemiInfinitePressureBand(is_source=True, is_on_body=False, parent=self)
            )
            self._end_pts[0] = -1000.0  # doesn't matter where
        else:
            self.pressure_elements.append(
                pe.ForwardHalfTriangularPressureElement(
                    is_source=True, is_on_body=False, parent=self
                )
            )

            self.smoothing_factor = smoothing_factor
            # TODO: Are we adding these elements twice?
            for n in [self.neighbor_down, self.neighbor_up]:
                if n is None and ~np.isnan(self.smoothing_factor):
                    self.pressure_elements.extend(
                        [
                            pe.CompleteTriangularPressureElement(
                                is_source=True, is_on_body=False, parent=self
                            )
                            for _ in range(num_elements)
                        ]
                    )

            if downstream_planing_surface is None:
                self._end_pts[0] = (
                    downstream_loc
                    if downstream_loc is not None
                    else getattr(self.config, "downstream_loc", 0.0)
                )

            if upstream_planing_surface is None:
                self._end_pts[1] = (
                    upstream_loc
                    if upstream_loc is not None
                    else getattr(self.config, "upstream_loc", 0.0)
                )

            self.pressure_elements.append(
                pe.AftHalfTriangularPressureElement(is_source=True, is_on_body=False, parent=self)
            )
        self.update_end_pts()

    @property
    def base_pt(self) -> float:
        """The x-coordinate of the base point at the upstream edge of the pressure cushion."""
        if self.neighbor_up is None:
            return self._end_pts[1]
        else:
            return self.neighbor_up.base_pt

    @base_pt.setter
    def base_pt(self, x: float) -> None:
        self._end_pts[1] = x

    @property
    def length(self) -> float:
        """Length of pressure cushion."""
        return super().length

    @length.setter
    def length(self, length: float) -> None:
        """Set the length by moving the left end point relative to the right one."""
        self._end_pts[0] = self.base_pt - length

    @property
    def element_coords(self) -> np.ndarray:
        """Return x-locations of all elements."""
        if self.cushion_type != "smoothed" or np.isnan(self.smoothing_factor):
            return self._end_pts
        else:
            add_width = np.arctanh(0.99) * self.length / (2 * self.smoothing_factor)
            dx = np.linspace(-add_width, add_width, len(self.pressure_elements) // 2 + 1)
            return np.hstack((self._end_pts[0] + dx, self._end_pts[1] + dx))

    def update_end_pts(self) -> None:
        """Update the end coordinates."""
        if self.neighbor_down is not None:
            self._end_pts[0] = self.neighbor_down._end_pts[1]
        if self.neighbor_up is not None:
            self._end_pts[1] = self.neighbor_up._end_pts[0]

        self._reset_element_coords()

        # Set source pressure of elements
        for el_num, el in enumerate(self.pressure_elements):
            el.is_source = True
            if self.cushion_type == "smoothed" and not np.isnan(self.smoothing_factor):
                alf = 2 * self.smoothing_factor / self.length
                el.pressure = (
                    0.5
                    * self.cushion_pressure
                    * (np.tanh(alf * el.x_coord) - np.tanh(alf * (el.x_coord - self.length)))
                )
            # for infinite pressure cushion, first element is dummy, set to
            # zero, second is semiInfinitePressureBand and set to cushion
            # pressure
            elif self.cushion_type == "infinite":
                el.pressure = 0.0 if el_num == 0 else self.cushion_pressure
            else:
                el.pressure = self.cushion_pressure


class PlaningSurface(PressurePatch):
    """Planing Surface consisting of unknown elements.

    Attributes:
        name: The name of the planing surface (should be unique, but that is unenforced).
        initial_length: An initial guessed wetted length of the planing surface.
        minimum_length: The minimum length of the surface allowed to be wetted.
        maximum_length: The maximum length of the surface allowed to be wetted.
        num_fluid_elements: The number of elements to discretize surface into.
        point_spacing: Method to use for distributing pressure elements. Either "linear" or "cosine".
        kutta_pressure: The target pressure at the trailing edge.
        upstream_pressure: The target pressure upstream of the stagnation point.
        is_sprung: If true, contribution due to numerical spring is added to the lift force in order to ensure
            the tip of the surface lies on the free surface.
        spring_constant: The spring constant.
        parent: The solver to which this planing surface belongs.

    """

    def __init__(
        self,
        *,
        name: str = "",
        initial_length: float = 0.1,
        minimum_length: float = 0.0,
        maximum_length: float = float("Inf"),
        num_fluid_elements: int = 1,
        point_spacing: Literal["linear"] | Literal["cosine"] = "linear",
        kutta_pressure: float | str = 0.0,
        upstream_pressure: float = 0.0,
        is_sprung: bool = False,
        spring_constant: float = 1e4,
        parent: PotentialPlaningSolver | None = None,
        **_: Any,
    ) -> None:
        super().__init__(parent=parent)

        self.name = name
        self.initial_length = initial_length
        self.minimum_length = minimum_length
        self.maximum_length = maximum_length

        self.kutta_pressure = kutta_pressure
        if isinstance(self.kutta_pressure, str):
            self.kutta_pressure = getattr(self.config.body, self.kutta_pressure)

        self._upstream_pressure = upstream_pressure
        if isinstance(self._upstream_pressure, str):
            self._upstream_pressure = getattr(self.config, self._upstream_pressure)

        self.is_initialized = False
        self.is_active = True
        self.is_kutta_unknown = True

        self.is_sprung = is_sprung
        self.spring_constant = spring_constant
        if self.is_sprung:
            self.initial_length = 0.0
            self.minimum_length = 0.0
            self.maximum_length = 0.0

        self.pressure_elements.append(
            pe.ForwardHalfTriangularPressureElement(parent=self, is_on_body=False)
        )
        self.pressure_elements.append(
            pe.ForwardHalfTriangularPressureElement(
                parent=self, is_source=True, pressure=self.kutta_pressure
            )
        )
        self.pressure_elements.extend(
            [
                pe.CompleteTriangularPressureElement(parent=self)
                for _ in range(num_fluid_elements - 1)
            ]
        )
        self.pressure_elements.append(
            pe.AftHalfTriangularPressureElement(
                parent=self, is_source=True, pressure=self.upstream_pressure
            )
        )

        # Define point spacing
        self.relative_position: np.ndarray
        if point_spacing == "cosine":
            # TODO: Remove ignore
            self.relative_position = 0.5 * (  # type: ignore
                1 - trig.cosd(np.linspace(0.0, 180.0, num_fluid_elements + 1))
            )
        elif point_spacing == "linear":
            self.relative_position = np.linspace(0.0, 1.0, num_fluid_elements + 1)
        else:
            raise ValueError("point_spacing must be 'cosine' or 'linear'")

        self.relative_position /= self.relative_position[-2]
        self.relative_position = np.hstack((0.0, self.relative_position))

        self.x_coords = np.array([])
        self.pressure = np.array([])
        self.shear_stress = np.array([])
        self.s_coords = np.array([])

    @property
    def upstream_pressure(self) -> float:
        """The pressure at the upstream end of the planing surface."""
        if isinstance(self.neighbor_up, PressureCushion):
            return self.neighbor_up.cushion_pressure
        else:
            return 0.0

    @upstream_pressure.setter
    def upstream_pressure(self, pressure: float) -> None:
        """When setting the upstream pressure, that element is now a source."""
        self._upstream_pressure = pressure
        el = self.pressure_elements[-1]
        el.pressure = pressure
        el.is_source = True
        el.z_coord = np.nan

    @property
    def downstream_pressure(self) -> float:
        """The pressure at the downstream end of the planing surface."""
        if isinstance(self.neighbor_down, PressureCushion):
            return self.neighbor_down.cushion_pressure
        else:
            return 0.0

    @property
    def length(self) -> float:
        """Length of planing surface."""
        return super().length

    @length.setter
    def length(self, length: float) -> None:
        """Set the length and re-distribute the elements with the base at the separation point."""
        length = np.min([np.max([length, 0.0]), self.maximum_length])
        assert self.interpolator is not None
        x0 = self.interpolator.get_separation_point()[0]
        self._end_pts[:] = [x0, x0 + length]
        self._reset_element_coords()

    @property
    def residual(self) -> float:
        """The residual to drive first element pressure to zero."""
        if self.length > 0.0:
            return self.pressure_elements[0].pressure / self.config.flow.stagnation_pressure
        else:
            return 0.0

    def initialize_end_pts(self) -> None:
        """Initialize end points to be at separation point and wetted length root."""
        assert self.interpolator is not None
        self.base_pt = self.interpolator.get_separation_point()[0]

        if not self.is_initialized:
            if self.initial_length is None:
                self.length = self.interpolator.immersed_length - self.base_pt
            else:
                self.length = self.initial_length
            self.is_initialized = True

    def _reset_element_coords(self) -> None:
        """Set width of first element to be twice as wide."""
        super()._reset_element_coords()
        x = self.element_coords
        self.pressure_elements[0].width = x[2] - x[0]

    @property
    def element_coords(self) -> np.ndarray:
        """Get position of pressure elements."""
        return self.base_pt + self.relative_position * self.length

    def calculate_forces(self) -> None:
        """Calculate the forces by integrating pressure and shear stress."""
        if self.config.flow.include_friction:
            self._calculate_shear_stress()

        if self.length > 0.0:
            el = [el for el in self.pressure_elements if el.is_on_body]
            self.x_coords = np.array([eli.x_coord for eli in el])
            self.pressure = np.array([eli.pressure for eli in el])
            self.pressure += self.pressure_elements[0].pressure
            self.shear_stress = np.array([eli.shear_stress for eli in el])
            assert self.interpolator is not None
            self.s_coords = np.array([self.interpolator.get_s_fixed_x(xx) for xx in self.x_coords])

            tangent_angle = trig.atand(self._get_body_derivative(self.x_coords))
            self.drag_pressure = math_helpers.integrate(
                self.s_coords, self.pressure * trig.sind(tangent_angle)
            )
            self.drag_friction = math_helpers.integrate(
                self.s_coords, self.shear_stress * trig.cosd(tangent_angle)
            )
            self.drag_total = self.drag_pressure + self.drag_friction
            self.lift_pressure = math_helpers.integrate(
                self.s_coords, self.pressure * trig.cosd(tangent_angle)
            )
            self.lift_friction = -math_helpers.integrate(
                self.s_coords, self.shear_stress * trig.sind(tangent_angle)
            )
            self.lift_total = self.lift_pressure + self.lift_friction
            self.moment_total = math_helpers.integrate(
                self.x_coords,
                # TODO: This shouldn't reference body config
                self.pressure * trig.cosd(tangent_angle) * (self.x_coords - self.config.body.xCofR),
            )
            if self.is_sprung:
                self._apply_spring()
        else:
            self.drag_pressure = 0.0
            self.drag_friction = 0.0
            self.drag_total = 0.0
            self.lift_pressure = 0.0
            self.lift_friction = 0.0
            self.lift_total = 0.0
            self.moment_total = 0.0
            self.x_coords = np.array([])

    def get_loads_in_range(
        self, x0: float, x1: float, /, *, pressure_limit: float | None = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Return pressure and shear stress values at points between two
        arguments.

        Args:
            x0: The starting x-coordinate.
            x1: The ending x-coordinate.
            pressure_limit: Optionally limit the pressure to be below a specified value (e.g. stagnation pressure).
        """
        if self.length > 0.0:
            ind = np.nonzero((self.x_coords > x0) * (self.x_coords < x1))[0]
            interp_p = interp1d(self.x_coords, self.pressure, bounds_error=False, fill_value=0.0)
            interp_tau = interp1d(
                self.x_coords, self.shear_stress, bounds_error=False, fill_value=0.0
            )

            x = np.hstack((x0, self.x_coords[ind], x1))
            p = np.hstack((interp_p(x0), self.pressure[ind], interp_p(x1)))
            tau = np.hstack((interp_tau(x0), self.shear_stress[ind], interp_tau(x1)))
        else:
            x = np.array([x0, x1])
            p = np.zeros_like(x)
            tau = np.zeros_like(x)

        if pressure_limit is not None:
            # Limit pressure to be below stagnation pressure
            p = np.min(
                np.vstack(
                    (
                        p,
                        np.ones_like(p) * pressure_limit,
                    )
                ),
                axis=0,
            )
        return x, p, tau

    def _apply_spring(self) -> None:
        """Apply the spring force to the total lift and moment forces."""
        x_s = self.pressure_elements[0].x_coord
        z_s = self.pressure_elements[0].z_coord
        assert self.parent is not None
        spring_displacement = z_s - self.parent.get_free_surface_height(x_s)
        spring_force = -self.spring_constant * spring_displacement
        self.lift_total += spring_force
        # TODO: This shouldn't reference body config
        self.moment_total += spring_force * (x_s - self.config.body.xCofR)

    def _calculate_shear_stress(self) -> None:
        """Calculate the shear stress and apply to each element."""

        def get_shear_stress(xx: float) -> float:
            """Calculate the shear stress at a given location."""
            re_x = self.config.flow.flow_speed * xx / self.config.flow.kinematic_viscosity
            return (
                0.332 * self.config.flow.density * self.config.flow.flow_speed**2 * re_x**-0.5
            )

        x = self.element_coords[:-1]
        assert self.interpolator is not None
        s = np.array([self.interpolator.get_s_fixed_x(xx) for xx in x])
        s = s[-1] - s

        for s_i, el in zip(s, self.pressure_elements):
            el.shear_stress = get_shear_stress(s_i) if s_i > 0.0 else 0.0

    def _get_body_derivative(self, x: np.ndarray, direction: str = "r") -> np.ndarray:
        """Calculate the derivative of the body surface at a point."""
        assert self.interpolator is not None
        return np.array(
            [math_helpers.deriv(self.interpolator.get_body_height, xx, direction) for xx in x]
        )



from __future__ import annotations

import json
import os
import re
from pathlib import Path
from re import Match
from typing import Any

from planingfsi import logger
from planingfsi import unit  # noqa: F401

__all__ = ["load_dict_from_file"]


def replace_single_quotes_with_double_quotes(string: str) -> str:
    """Replace all single-quoted strings with double-quotes."""

    def repl(m: Match) -> str:
        return m.group(1).join('""')

    return re.sub(r"'(.+?)'", repl, string)


def replace_environment_variables(string: str) -> str:
    """Replace environment variables with their value."""

    def repl(m: Match) -> str:
        return os.environ[m.group(1)]

    return re.sub(r"\$(\w+)", repl, string)


def add_quotes_to_words(string: str) -> str:
    """Find words inside a string and surround with double-quotes."""
    quoted_pattern = re.compile('(".+?")')
    word_pattern = re.compile(r"([\w.-]+)")
    # Any number, integer, float, or exponential
    number_pattern = re.compile(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?")

    matches = quoted_pattern.split(string)

    def repl(m: Match) -> str:
        """Add quotes, only if it isn't a number."""
        value = m.group(1)
        if number_pattern.match(value):
            return value
        return value.join('""')

    return "".join(word_pattern.sub(repl, m) if i % 2 == 0 else m for i, m in enumerate(matches))


def jsonify_string(string: str) -> str:
    """Loop through a string, ensuring double-quotes are used to comply with json standard.

    * Find pattern.
    * Add everything up until the pattern to new copy of string.
    * Add pattern with single-quotes substituted for double-quotes.
    * If there's a double-quote inside single-quotes, it means we have an apostrophe.
    * Add everything up to the double quote.
    * Once they are all added, which will eventually include the apostrophe, normal matching will
      proceed.

    """
    if not string.startswith("{"):
        string = string.join("{}")

    string = replace_single_quotes_with_double_quotes(string)
    string = add_quotes_to_words(string)
    string = replace_environment_variables(string)

    string = re.sub(",+", ",", string)
    string = (
        string.replace("[,", "[")
        .replace("{,", "{")
        .replace(",]", "]")
        .replace(",}", "}")
        .replace("}{", "},{")
    )

    logger.debug(f'JSONified string: "{string}"')

    return string


def load_dict_from_file(
    filename: Path | str, key_map: dict[str, str] | None = None
) -> dict[str, Any]:
    """Read a file, which is a less strict JSON format, and return a dictionary.

    Optionally, a key map may be provided to allow loading older files by
    replacing keys with updated spellings. For example, a `key_map = {"oldKey": "old_key"}`
    could read a file containing the key "oldKey", but the dictionary that is
    returned will have replaced that key with "old_key". When using the `key_map`
    functionality, an exception will be raised if both the old and new keys exist
    in the dictionary being loaded.

    Args:
        filename: A filename or path to the file to be loaded.
        key_map: An optional mapping of keys.

    Returns:
        A dictionary mapping keys to values from the input file.

    """
    logger.debug('Loading Dictionary from file "{}"'.format(filename))

    with Path(filename).open() as f:
        dict_iter = (line.split("#")[0].strip() for line in f.readlines())
    try:
        dict_ = load_dict_from_string(",".join(dict_iter))
    except ValueError:
        print(f"Error reading file {filename}")
        raise

    # If specified, read values from a base dictionary
    # All local values override the base dictionary values
    base_dict_dir = dict_.get("baseDict", dict_.get("base_dict"))
    if base_dict_dir:
        base_dict_dir = os.path.split(base_dict_dir)
        # Tracing relative references from original file directory
        if base_dict_dir[0].startswith("."):
            base_dict_dir = os.path.abspath(os.path.join(os.path.dirname(filename), *base_dict_dir))
        base_dict = load_dict_from_file(base_dict_dir)
        dict_.update({k: v for k, v in base_dict.items() if k not in dict_})

    if key_map:
        dict_ = _apply_key_map(dict_, key_map)

    return dict_


def _apply_key_map(dict_: dict[str, Any], key_map: dict[str, str]) -> dict[str, Any]:
    """Map old keys to new keys."""
    for old_key, new_key in key_map.items():
        if old_key in dict_ and new_key in dict_:
            raise KeyError(f"Cannot use both '{old_key}' and '{new_key}'")
        if old_key in dict_:
            dict_[new_key] = dict_.pop(old_key)
    return dict_


def load_dict_from_string(string: str) -> dict[str, Any]:
    """Convert string to JSON string, convert to a dictionary, and return."""
    logger.debug('Loading Dictionary from string: "{}"'.format(string))

    json_string = jsonify_string(string)
    try:
        dict_ = json.loads(json_string)
    except json.decoder.JSONDecodeError:
        raise ValueError('Error loading from json string: "{}"'.format(json_string))

    # Provide specialized handling of certain strings
    for key, val in dict_.items():
        if isinstance(val, str):
            match = re.match(r"([+-]?nan|[+-]?inf)", val)
            if match:
                dict_[key] = float(match.group(1))
            elif "unit." in val:
                dict_[key] = eval(val)

    return dict_



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np

from planingfsi import logger
from planingfsi.config import NUM_DIM
from planingfsi.config import Config
from planingfsi.fe.felib import Node
from planingfsi.fe.femesh import Mesh
from planingfsi.fe.rigid_body import RigidBody
from planingfsi.fe.substructure import Substructure

if TYPE_CHECKING:
    from planingfsi.simulation import Simulation


class StructuralSolver:
    """High-level finite-element structural solver, consisting of a number of rigid bodies.

    Attributes:
        simulation: A reference to the parent `Simulation` object.
        rigid_bodies: A list of all `RigidBody` instances in the simulation.
        nodes: A list of all nodes in the mesh.
        node_dofs: A mapping of node to the indices for that node in the global matrices.

    """

    def __init__(self, simulation: Simulation):
        self.simulation = simulation
        self.rigid_bodies: list[RigidBody] = []
        self.nodes: list[Node] = []
        self.node_dofs: dict[Node, list[int]] = {}

    @property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        return self.simulation.config

    @property
    def has_free_structure(self) -> bool:
        """True if any rigid body or substructure are free to move."""
        for rigid_body in self.rigid_bodies:
            if rigid_body.has_free_dof:
                return True

            for ss in rigid_body.substructures:
                if ss.is_free:
                    return True
        return False

    @property
    def residual(self) -> float:
        """The current maximum residual amongst all rigid bodies in the simulation."""
        return max((bd.residual for bd in self.rigid_bodies), default=0.0)

    @property
    def substructures(self) -> list[Substructure]:
        """A combined list of all substructures from all rigid bodies."""
        return [ss for body in self.rigid_bodies for ss in body.substructures]

    def add_rigid_body(
        self, dict_or_instance: dict[str, Any] | RigidBody | None = None
    ) -> RigidBody:
        """Add a rigid body to the structure.

        Args:
            dict_or_instance: A dictionary of values, or a RigidBody instance.

        """
        if isinstance(dict_or_instance, RigidBody):
            rigid_body = dict_or_instance
            rigid_body.parent = self
        else:
            dict_ = dict_or_instance or {}
            rigid_body = RigidBody(**dict_, parent=self)
        self.rigid_bodies.append(rigid_body)
        return rigid_body

    def add_substructure(
        self, dict_or_instance: dict[str, Any] | Substructure | None = None
    ) -> "Substructure":
        """Add a substructure to the structure, whose type is determined at run-time.

        Args:
            dict_or_instance: A dictionary of values, or a Substructure instance.

        """
        if isinstance(dict_or_instance, Substructure):
            ss = dict_or_instance
            dict_ = {}
        else:
            dict_ = dict_or_instance or {}
            ss = Substructure(type=dict_.get("substructureType", "rigid"), **dict_)
        ss.solver = self
        self.substructures.append(ss)

        self._assign_substructure_to_body(ss, **dict_)

        return ss

    def _assign_substructure_to_body(
        self, ss: "Substructure", body_name: str = "default", **_: Any
    ) -> None:
        """Find parent body and add substructure to it."""
        bodies = [b for b in self.rigid_bodies if b.name == body_name]
        if bodies:
            body = bodies[0]
        else:
            body = self.rigid_bodies[0]
        body.add_substructure(ss)
        logger.info(
            f"Adding Substructure {ss.name} of type {type(ss).__name__} to rigid body {body.name}"
        )

    def initialize_rigid_bodies(self) -> None:
        """Initialize the position of all rigid bodies."""
        for bd in self.rigid_bodies:
            bd.initialize_position()

    def update_fluid_forces(self) -> None:
        """Update fluid forces on all rigid bodies."""
        for bd in self.rigid_bodies:
            bd.update_fluid_forces()

    def calculate_response(self) -> None:
        """Calculate the structural response, or load from files."""
        if self.config.io.results_from_file:
            self._load_response()
        else:
            for bd in self.rigid_bodies:
                bd.update_position()
                bd.update_substructure_positions()

    def _load_response(self) -> None:
        """Load the response from files."""
        self.update_fluid_forces()

        for bd in self.rigid_bodies:
            bd.load_motion()
            for ss in bd.substructures:
                ss.load_coordinates()

    def write_results(self) -> None:
        """Write the results to file."""
        for bd in self.rigid_bodies:
            bd.write_motion()
            for ss in bd.substructures:
                ss.write_coordinates()

    def _load_mesh_from_object(self, mesh: Mesh) -> None:
        """Load a mesh from an existing object."""
        for pt in mesh.points:
            nd = Node(
                coordinates=pt.position, is_dof_fixed=pt.is_dof_fixed, fixed_load=pt.fixed_load
            )
            node_num = len(self.nodes)  # Needs to be before we append the node
            self.nodes.append(nd)
            self.node_dofs[nd] = [node_num * NUM_DIM + i for i in [0, 1]]

        for struct in self.substructures:
            # Find submesh with same name as substructure
            submesh = [submesh for submesh in mesh.submesh if submesh.name == struct.name][0]
            struct.load_mesh(submesh)

    def _load_mesh_from_dir(self, mesh_dir: Path) -> None:
        """Load the mesh from a directory of files."""
        # TODO: Can we do Mesh.from_dir() instead and then integrate the repeated logic into _load_mesh_from_object?
        coords = np.loadtxt(mesh_dir / "nodes.txt")
        fixed_dofs = np.loadtxt(mesh_dir / "fixedDOF.txt")
        loads = np.loadtxt(mesh_dir / "fixedLoad.txt")
        for c, fixed_dof, load in zip(coords, fixed_dofs, loads):
            nd = Node(coordinates=c, is_dof_fixed=fixed_dof, fixed_load=load)
            node_num = len(self.nodes)  # Needs to be before we append the node
            self.nodes.append(nd)
            self.node_dofs[nd] = [node_num * NUM_DIM + i for i in [0, 1]]

        for struct in self.substructures:
            struct.load_mesh(mesh_dir)

    def load_mesh(self, mesh: Path | Mesh = Path("mesh")) -> None:
        """Load the mesh from a directory of files or an existing mesh object."""
        # Create all nodes
        if isinstance(mesh, Mesh):
            self._load_mesh_from_object(mesh)
        else:
            self._load_mesh_from_dir(mesh)



"""Module containing definitions of different types of pressure element."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from scipy.special import sici

from planingfsi import math_helpers
from planingfsi.config import Config

if TYPE_CHECKING:
    from planingfsi.potentialflow.pressurepatch import PressurePatch


def _get_aux_fg(lam: float) -> tuple[float, float]:
    """Return f and g functions, which are dependent on the auxiliary
    sine and cosine functions. Combined into one function to reduce number of
    calls to scipy.special.sici().
    """
    lam = abs(lam)
    (aux_sine, aux_cosine) = sici(lam)
    (sine, cosine) = np.sin(lam), np.cos(lam)
    return (
        cosine * (0.5 * np.pi - aux_sine) + sine * aux_cosine,
        sine * (0.5 * np.pi - aux_sine) - cosine * aux_cosine,
    )


def _get_gamma1(lam: float, aux_g: float) -> float:
    """Return result of first integral (Gamma_1 in my thesis).

    Args:
        lam: Non-dimensional x-position.
        aux_g: Second result of `_get_aux_fg`.

    """
    if lam > 0:
        return (aux_g + np.log(lam)) / np.pi
    else:
        return (aux_g + np.log(-lam)) / np.pi + (2 * np.sin(lam) - lam)


def _get_gamma2(lam: float, aux_f: float) -> float:
    """Return result of second integral (Gamma_2 in my thesis).

    Args
    ----
    lam : float
        Non-dimensional x-position.
    aux_f : float
        First result of `_get_aux_fg`

    Returns
    -------
    float

    """
    return math_helpers.sign(lam) * aux_f / np.pi + math_helpers.heaviside(-lam) * (
        2 * np.cos(lam) - 1
    )


def _get_gamma3(lam: float, aux_f: float) -> float:
    """Return result of third integral (Gamma_3 in my thesis).

    Args
    ----
    lam : float
        Non-dimensional x-position.
    aux_f : float
        First result of `_get_aux_fg`

    Returns
    -------
    float
    """
    return (
        -math_helpers.sign(lam) * aux_f / np.pi
        - 2 * math_helpers.heaviside(-lam) * np.cos(lam)
        - math_helpers.heaviside(lam)
    )


class PressureElement(abc.ABC):
    """Abstract base class to represent all different types of pressure elements.

    Attributes:
        x_coord: x-location of element.
        z_coord: z-coordinate of element, if on body.
        pressure: The pressure/strength of the element.
        shear_stress: Shear stress at the element.
        width: Width of element.
        is_source: True of element is a source.
        is_on_body: True if on body. Used for force calculation.
        parent: Pressure patch that this element belongs to.

    """

    plot_color = "b"

    def __init__(
        self,
        x_coord: float = np.nan,
        z_coord: float = np.nan,
        pressure: float = np.nan,
        shear_stress: float = 0.0,
        width: np.ndarray | float = np.nan,
        is_source: bool = False,
        is_on_body: bool = False,
        parent: PressurePatch | None = None,
    ):
        # TODO: Replace x_coord with coord pointer to Coordinate object, (e.g. self.coords.x)
        self.x_coord = x_coord
        self.z_coord = z_coord
        self.pressure = pressure
        self.shear_stress = shear_stress
        self._width = np.zeros(2)
        # TODO: Remove ignore
        self.width = width  # type: ignore
        self.is_source: bool = is_source
        self.is_on_body: bool = is_on_body
        self.parent = parent

    @property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        if self.parent is None:
            raise ValueError("Must set parent pressure patch to access config.")
        return self.parent.config

    @property
    def width(self) -> float:
        """The total width of the pressure element."""
        return self._width.sum()

    @width.setter
    def width(self, width: float) -> None:
        self._width[0] = width

    def get_influence_coefficient(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient coefficient of element.

        Args:
            x_coord: Dimensional x-coordinate.

        """
        x_rel = x_coord - self.x_coord
        if self.width == 0.0:
            return 0.0
        elif x_rel in [-self._width[0], 0.0, self._width[1]]:
            dx = 1e-6
            influence = 0.5 * (
                self._get_local_influence_coefficient(x_rel + dx)
                + self._get_local_influence_coefficient(x_rel - dx)
            )
        else:
            influence = self._get_local_influence_coefficient(x_rel)
        return influence / (self.config.flow.density * self.config.flow.gravity)

    def get_influence(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient for actual pressure.

        Args:
            x_coord: Dimensional x-coordinate.

        """
        return self.get_influence_coefficient(x_coord) * self.pressure

    @abc.abstractmethod
    def _get_local_influence_coefficient(self, x_coord: float) -> float:
        """Return influence coefficient in iso-geometric coordinates.

        Args:
            x_coord: x-coordinate, centered at element location.

        """
        pass

    def __repr__(self) -> str:
        """Print element attributes."""
        return ", ".join(
            [
                f"{self.__class__.__name__}: (x,z) = ({self.x_coord}, {self.z_coord})",
                f"width = {self.width}",
                f"is_source = {self.is_source}",
                f"p = {self.pressure}",
                f"is_on_body = {self.is_on_body}",
            ]
        )

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return np.array([]), np.array([])


class AftHalfTriangularPressureElement(PressureElement):
    """Pressure element that is triangular in profile but towards aft direction."""

    plot_color = "g"

    def __init__(self, is_on_body: bool = True, **kwargs: Any) -> None:
        # By default, this element is on the body.
        super().__init__(is_on_body=is_on_body, **kwargs)

    @property
    def width(self) -> float:
        """The total width of the pressure element."""
        return super().width

    @width.setter
    def width(self, width: float) -> None:
        self._width[0] = width

    def _get_local_influence_coefficient(self, x_rel: float) -> float:
        """Get influence coefficient for element relative to the element position."""
        lambda_0 = self.config.flow.k0 * x_rel
        aux_f, aux_g = _get_aux_fg(lambda_0)
        influence = _get_gamma1(lambda_0, aux_g) / (self.width * self.config.flow.k0)
        influence += _get_gamma2(lambda_0, aux_f)

        lambda_2 = self.config.flow.k0 * (x_rel + self.width)
        _, aux_g = _get_aux_fg(lambda_2)
        influence -= _get_gamma1(lambda_2, aux_g) / (self.width * self.config.flow.k0)
        return influence

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return self.x_coord - np.array([self.width, 0]), np.array([0.0, self.pressure])


class ForwardHalfTriangularPressureElement(PressureElement):
    """Pressure element that is triangular in profile but towards forward direction."""

    def __init__(self, is_on_body: bool = True, **kwargs: Any) -> None:
        # By default, this element is on the body.
        super().__init__(is_on_body=is_on_body, **kwargs)

    @property
    def width(self) -> float:
        """The total width of the pressure element."""
        return super().width

    @width.setter
    def width(self, width: float) -> None:
        self._width[1] = width

    def _get_local_influence_coefficient(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient coefficient in iso-geometric coordinates.

        Args:
            x_coord: x-coordinate, centered at element location.

        """
        lambda_0 = self.config.flow.k0 * x_coord
        aux_f, aux_g = _get_aux_fg(lambda_0)
        influence = _get_gamma1(lambda_0, aux_g) / (self.width * self.config.flow.k0)
        influence -= _get_gamma2(lambda_0, aux_f)

        lambda_1 = self.config.flow.k0 * (x_coord - self.width)
        _, aux_g = _get_aux_fg(lambda_1)
        influence -= _get_gamma1(lambda_1, aux_g) / (self.width * self.config.flow.k0)
        return influence

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return self.x_coord + np.array([0.0, self.width]), np.array([self.pressure, 0.0])


class CompleteTriangularPressureElement(PressureElement):
    """Pressure element that is triangular in profile towards both directions."""

    plot_color = "r"

    def __init__(
        self, is_on_body: bool = True, width: np.ndarray = np.zeros(2), **kwargs: Any
    ) -> None:
        # By default, this element is on the body.
        super().__init__(is_on_body=is_on_body, width=width, **kwargs)

    @property
    def width(self) -> float:
        """The total width of the pressure element."""
        return super().width

    @width.setter
    def width(self, width: np.ndarray) -> None:
        if not len(width) == 2:
            raise ValueError("Width must be length-two array")
        else:
            self._width = np.array(width)

    def _get_local_influence_coefficient(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient coefficient in iso-geometric
        coordinates.

        Args
        ----
        x_coord : float
            x-coordinate, centered at element location.
        """
        lambda_0 = self.config.flow.k0 * x_coord
        _, aux_g = _get_aux_fg(lambda_0)
        influence = _get_gamma1(lambda_0, aux_g) * self.width / (self._width[1] * self._width[0])

        lambda_1 = self.config.flow.k0 * (x_coord - self._width[1])
        _, aux_g = _get_aux_fg(lambda_1)
        influence -= _get_gamma1(lambda_1, aux_g) / self._width[1]

        lambda_2 = self.config.flow.k0 * (x_coord + self._width[0])
        _, aux_g = _get_aux_fg(lambda_2)
        influence -= _get_gamma1(lambda_2, aux_g) / self._width[0]
        return influence / self.config.flow.k0

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return (
            self.x_coord + np.array([-self._width[0], 0.0, self._width[1]]),
            np.array([0.0, self.pressure, 0.0]),
        )


class AftSemiInfinitePressureBand(PressureElement):
    """Semi-infinite pressure band in aft direction."""

    def _get_local_influence_coefficient(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient coefficient in
        iso-geometric coordinates.

        Args
        ----
        x_coord : float
            x-coordinate, centered at element location.
        """
        lambda_0 = self.config.flow.k0 * x_coord
        aux_f, _ = _get_aux_fg(lambda_0)
        influence = _get_gamma2(lambda_0, aux_f)
        return influence

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return (
            np.array([self.config.plotting.x_fs_min, self.x_coord]),
            np.array([self.pressure, self.pressure]),
        )


class ForwardSemiInfinitePressureBand(PressureElement):
    """Semi-infinite pressure band in forward direction."""

    plot_color = "r"

    def _get_local_influence_coefficient(self, x_coord: float) -> float:
        """Return _get_local_influence_coefficient coefficient in iso-geometric coordinates.

        Args:
            x_coord: x-coordinate, centered at element location.

        """
        lambda_0 = self.config.flow.k0 * x_coord
        aux_f, _ = _get_aux_fg(lambda_0)
        influence = _get_gamma3(lambda_0, aux_f)
        return influence

    @property
    def plot_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Coordinates for pressure plot."""
        return (
            np.array([self.x_coord, self.config.plotting.x_fs_max]),
            np.array([self.pressure, self.pressure]),
        )



"""High-level control of a `planingfsi` simulation."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np

from planingfsi import Mesh
from planingfsi import logger
from planingfsi import writers
from planingfsi.config import Config
from planingfsi.dictionary import load_dict_from_file
from planingfsi.fe.structure import StructuralSolver
from planingfsi.figure import Figure
from planingfsi.potentialflow.pressurepatch import PlaningSurface
from planingfsi.potentialflow.solver import PotentialPlaningSolver

if TYPE_CHECKING:
    from planingfsi.fe.rigid_body import RigidBody


class Simulation:
    """Simulation object to manage the FSI problem.

    Handles the iteration between the fluid and solid solvers.

    Attributes:
        config: A reference to the `Config` object, which stores values "global" to this simulation.
            Each `Simulation` instance has its own independent config.
        structural_solver: The structural solver.
        fluid_solver: The fluid solver.
        it: The current iteration.
        ramp: The current ramping coefficient, used when applying loads and moving nodes.
            The value is between 1.0 and 0.0.

    """

    def __init__(self) -> None:
        self.config = Config()
        self.structural_solver = StructuralSolver(self)
        self.fluid_solver = PotentialPlaningSolver(self)
        self.it = 0
        self.ramp = 1.0
        self._figure: Figure | None = None
        self._it_dirs: list[Path] | None = None
        self._case_dir: Path | None = None

    @classmethod
    def from_input_files(cls, config_filename: Path | str) -> Simulation:
        """Construct a `Simulation` object by loading input files."""
        simulation = cls()
        simulation.load_input_files(config_filename)
        return simulation

    @property
    def case_dir(self) -> Path:
        """The base path for the simulation."""
        return self._case_dir or Path(self.config.path.case_dir)

    @case_dir.setter
    def case_dir(self, value: Path) -> None:
        self._case_dir = value

    @property
    def mesh_dir(self) -> Path:
        """A path to the mesh directory."""
        return self.case_dir / self.config.path.mesh_dir_name

    @property
    def fig_dir(self):
        """A path to the directory in which figures will be saved."""
        return self.case_dir / self.config.path.fig_dir_name

    @property
    def figure(self) -> Figure | None:
        """The `FSIFigure` object where results are drawn. Will be None if plotting is disabled."""
        if self._figure is None and self.config.plotting.plot_any:
            self._figure = Figure(simulation=self)
        return self._figure

    @property
    def it_dir(self) -> Path:
        """A path to the directory for the current iteration."""
        return self.case_dir / str(self.it)

    @property
    def it_dirs(self) -> list[Path]:
        if self._it_dirs is None:
            self._it_dirs = sorted(self.case_dir.glob("[0-9]*"), key=lambda x: int(x.name))
        return self._it_dirs

    @property
    def residual(self) -> float:
        return self.structural_solver.residual

    @property
    def is_converged(self) -> bool:
        return (
            self.structural_solver.residual < self.config.solver.max_residual
            and self.it > self.config.solver.num_ramp_it
        )

    @property
    def is_write_iteration(self) -> bool:
        """True if results should be written at the end of the current iteration."""
        return (
            self.it >= self.config.solver.max_it
            or self.residual < self.config.solver.max_residual
            or np.mod(self.it, self.config.io.write_interval) == 0
        )

    def add_rigid_body(self, rigid_body: dict[str, Any] | RigidBody | None = None) -> RigidBody:
        """Add a rigid body to the simulation and solid solver.

        Args:
            rigid_body: A `RigidBody` object, or optional dictionary of values to construct the rigid body.

        Returns:
            The `RigidBody` that was added to the simulation.

        """
        return self.structural_solver.add_rigid_body(rigid_body)

    def _update_fluid_response(self) -> None:
        """Update the fluid response and apply to the structural solver."""
        self.fluid_solver.calculate_response()
        self.structural_solver.update_fluid_forces()

    def _update_solid_response(self) -> None:
        """Update the structural response."""
        self.structural_solver.calculate_response()

    def load_input_files(self, config_filename: Path | str) -> None:
        """Load all the input files."""
        self.config.load_from_file(config_filename)
        self._load_rigid_bodies()
        self._load_substructures()
        self._load_pressure_cushions()
        self.load_mesh()

    def _load_rigid_bodies(self) -> None:
        """Load all rigid bodies from files in the body dict directory.

        If no files are provided, a default stationary rigid body is added.

        """
        body_dict_dir = self.case_dir / self.config.path.body_dict_dir_name
        for dict_path in body_dict_dir.glob("*"):
            # TODO: I may be missing old spellings in the key_map
            dict_ = load_dict_from_file(
                dict_path,
                key_map={
                    "bodyName": "name",
                    "W": "weight",
                    "loadPct": "load_pct",
                    "m": "mass",
                    "Iz": "rotational_inertia",
                    "xCofG": "x_cg",
                    "yCofG": "y_cg",
                    "xCofR": "x_cr",
                    "yCofR": "y_cr",
                },
            )

            # Fallback on self.config.body if these keys are missing from rigid body dict file
            fallback_keys = {
                "x_cg",
                "y_cg",
                "x_cr",
                "y_cr",
                "initial_draft",
                "initial_trim",
                "free_in_draft",
                "free_in_trim",
                "max_draft_step",
                "max_trim_step",
                "relax_draft",
                "relax_trim",
            }
            for key in fallback_keys:
                dict_.setdefault(key, getattr(self.config.body, key))

            dict_.setdefault("weight", self.config.body.seal_load_pct * self.config.body.weight)

            self.add_rigid_body(dict_)

        if not self.structural_solver.rigid_bodies:
            # Add a dummy rigid body that cannot move
            self.add_rigid_body()

        logger.info(f"Rigid Bodies: {self.structural_solver.rigid_bodies}")

    def _load_substructures(self) -> None:
        """Load all substructures from files."""
        input_dict_dir = self.case_dir / self.config.path.input_dict_dir_name
        for dict_path in input_dict_dir.glob("*"):
            dict_ = load_dict_from_file(
                dict_path,
                key_map={
                    "substructureName": "name",
                    "initialLength": "initial_length",
                    "Nfl": "num_fluid_elements",
                    "minimumLength": "minimum_length",
                    "maximumLength": "maximum_length",
                    "kuttaPressure": "kutta_pressure",
                    "upstreamPressure": "upstream_pressure",
                    "isSprung": "is_sprung",
                    "springConstant": "spring_constant",
                    "pointSpacing": "point_spacing",
                    "hasPlaningSurface": "has_planing_surface",
                    "waterlineHeight": "waterline_height",
                    "sSepPctStart": "separation_arclength_start_pct",
                    "sImmPctStart": "immersion_arclength_start_pct",
                    "bodyName": "body_name",
                    "Ps": "seal_pressure",
                    "PsMethod": "seal_pressure_method",
                    "overPressurePct": "seal_over_pressure_pct",
                    "cushionPressureType": "cushion_pressure_type",
                    "tipLoad": "tip_load",
                    "tipLoadPct": "tip_load_pct",
                    "tipConstraintHt": "tip_constraint_height",
                    "structInterpType": "struct_interp_type",
                    "structExtrap": "struct_extrap",
                    "initialAngle": "initial_angle",
                    "basePtPct": "base_pt_pct",
                    "relaxAng": "relaxation_angle",
                    "attachPct": "attach_pct",
                    "minimumAngle": "minimum_angle",
                    "maxAngleStep": "max_angle_step",
                    "attachedSubstructure": "attached_substructure_name",
                    "attachedSubstructureEnd": "attached_substructure_end",
                    "EA": "axial_stiffness",
                },
            )
            # TODO: This default fallback to config could be handled in the PlaningSurface class
            dict_.setdefault("waterline_height", self.config.flow.waterline_height)

            substructure = self.structural_solver.add_substructure(dict_)

            if dict_.get("has_planing_surface", False):
                planing_surface = PlaningSurface(**dict_)
                substructure.add_planing_surface(planing_surface, **dict_)
        logger.info(f"Substructures: {self.structural_solver.substructures}")

    def _load_pressure_cushions(self) -> None:
        """Load all pressure cushions from files."""
        cushion_dict_dir = self.case_dir / self.config.path.cushion_dict_dir_name
        for dict_path in cushion_dict_dir.glob("*"):
            dict_ = load_dict_from_file(
                dict_path,
                key_map={
                    "pressureCushionName": "name",
                    "cushionPressure": "cushion_pressure",
                    "upstreamPlaningSurface": "upstream_planing_surface",
                    "downstreamPlaningSurface": "downstream_planing_surface",
                    "upstreamLoc": "upstream_loc",
                    "downstreamLoc": "downstream_loc",
                    "numElements": "num_elements",
                    "cushionType": "cushion_type",
                    "smoothingFactor": "smoothing_factor",
                },
            )
            self.fluid_solver.add_pressure_cushion(dict_)
        logger.info(f"Pressure Cushions: {self.fluid_solver.pressure_cushions}")

    def load_mesh(self, mesh: Path | Mesh | None = None) -> None:
        """Load the mesh from files, or directly. By default, will load from "mesh" directory."""
        self.structural_solver.load_mesh(mesh or self.mesh_dir)

    def run(self) -> None:
        """Run the fluid-structure interaction simulation.

        The fluid and solid solvers are solved iteratively until convergence is reached.

        """
        self.it = 0

        if self.config.io.results_from_file:
            self._update_ramp()

        self.initialize_solvers()

        # Iterate between solid and fluid solvers until equilibrium
        while self.it <= self.config.solver.num_ramp_it or (
            self.residual >= self.config.solver.max_residual
            and self.it <= self.config.solver.max_it
        ):

            # Calculate response
            if self.structural_solver.has_free_structure:
                self._update_ramp()
                self._update_solid_response()
                self._update_fluid_response()

            # Write, print, and plot results
            self.write_results()
            logger.info("Residual after iteration %4s: %5.3e", self.it, self.residual)
            self._update_figure()

            # Increment iteration count
            self.increment()

        self._save_figure()

        logger.info("Execution complete")

        if self.figure is not None and self.config.plotting.show:
            self.figure.show()

    def initialize_solvers(self) -> None:
        """Initialize body at specified trim and draft and solve initial fluid problem."""
        self.structural_solver.initialize_rigid_bodies()
        self._update_fluid_response()

    def increment(self) -> None:
        """Increment iteration counter. If loading from files, use the next stored iteration."""
        if self.config.io.results_from_file:
            stored_iterations = [int(it_dir.name) for it_dir in self.it_dirs]
            current_ind = stored_iterations.index(self.it)
            try:
                self.it = stored_iterations[current_ind + 1]
            except IndexError:
                self.it = self.config.solver.max_it + 1
        else:
            self.it += 1

    def _update_figure(self) -> None:
        if self.figure is not None and self.config.plotting.plot_any:
            self.figure.update()

    def _save_figure(self) -> None:
        """Save the final figure if configuration is correct."""
        if (
            self.figure is not None
            and self.config.plotting.save
            and not self.config.plotting.fig_format == "png"
        ):
            self.config.plotting.fig_format = "png"
            self.figure.save()

        if self.figure is not None and self.config.io.write_time_histories:
            self.figure.write_time_histories()

    def _update_ramp(self) -> None:
        """Update the ramp value based on the current iteration number."""
        if self.config.io.results_from_file:
            self._load_results()
        else:
            if self.config.solver.num_ramp_it == 0:
                self.ramp = 1.0
            else:
                self.ramp = min(self.it / self.config.solver.num_ramp_it, 1.0)

            self.config.solver.relax_FEM = (
                1 - self.ramp
            ) * self.config.solver.relax_initial + self.ramp * self.config.solver.relax_final

    def write_results(self) -> None:
        """Write the current overall results to an iteration directory."""
        if self.is_write_iteration and not self.config.io.results_from_file:
            self.it_dir.mkdir(exist_ok=True, parents=True)
            writers.write_as_dict(
                self.it_dir / "overallQuantities.txt",
                ["Ramp", self.ramp],
                ["Residual", self.structural_solver.residual],
            )

            self.fluid_solver.write_results()
            self.structural_solver.write_results()

    def _load_results(self) -> None:
        """Load the overall quantities from the results file."""
        dict_ = load_dict_from_file(self.it_dir / "overallQuantities.txt")
        self.ramp = dict_.get("Ramp", 0.0)
        self.structural_solver.residual = dict_.get("Residual", 0.0)



"""A simple module for storing unit conversion factors to convert to SI.

Example: A length e.g. can be represented as 1.0 * unit.ft to represent one foot,
    which will return 0.3048 meters.

"""
import math

# Acceleration due to gravity
_gravity = 9.80665

# Convert length to m
m = 1.0
cm = m / 100.0
mm = m / 1000.0
ft = 0.3048
inch = ft / 12.0

# Convert speed to m/s
mps = 1.0
fps = ft
kts = 0.5144444

# Convert mass to kg
kg = 1.0
lbm = kg / 2.20462

# Convert force to N
N = 1.0
lb = _gravity * lbm
lbf = 1.0 * lb  # lbs-force
mtf = _gravity * 1000.0  # metric tonnes-force
stf = N * lb * 2000.0

rad = 1.0
deg = math.pi / 180.0



from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np

from planingfsi import logger
from planingfsi import solver
from planingfsi import trig
from planingfsi import writers
from planingfsi.config import NUM_DIM
from planingfsi.config import Config
from planingfsi.dictionary import load_dict_from_file
from planingfsi.fe import felib as fe
from planingfsi.fe.substructure import FlexibleMembraneSubstructure
from planingfsi.fe.substructure import GlobalLoads
from planingfsi.fe.substructure import TorsionalSpringSubstructure
from planingfsi.solver import RootFinder

if TYPE_CHECKING:
    from planingfsi.fe.structure import StructuralSolver
    from planingfsi.fe.substructure import Substructure


class RigidBody:
    """A rigid body, consisting of multiple substructures.

    The rigid body is used as a container for collecting attached substructures, and also
    allows motion in heave & trim, which is calculated by integrating the total forces
    and moments applied to the substructures.

    Args:
        name: A name for the RigidBody instance
        weight: If provided, the weight to use for the rigid body (scaled by `config.body.seal_load_pct`).
        x_cg: The x-coordinate of the center of gravity.
        y_cg: The y-coordinate of the center of gravity.
        x_cr: The x-coordinate of the center of rotation. Defaults to the CG.
        y_cr: The y-coordinate of the center of rotation. Defaults to the CG.
        initial_draft: The initial draft of the body.
        initial_trim: The initial trim of the body.
        relax_draft: The relaxation parameter to use for draft.
        relax_trim: The relaxation parameter to use for trim.
        max_draft_step: The maximum allowable change in draft during an iteration.
        max_trim_step: The maximum allowable change in trim during an iteration.
        free_in_draft: True if body is free to move in draft.
        free_in_trim: True if body is free to move in trim.
        parent: An optional reference to the parent structural solver.

    """

    def __init__(
        self,
        name: str = "default",
        weight: float = 1.0,
        x_cg: float = 0.0,
        y_cg: float = 0.0,
        x_cr: float | None = None,
        y_cr: float | None = None,
        initial_draft: float = 0.0,
        initial_trim: float = 0.0,
        relax_draft: float = 1.0,
        relax_trim: float = 1.0,
        max_draft_step: float = 1e-3,
        max_trim_step: float = 1e-3,
        free_in_draft: bool = False,
        free_in_trim: bool = False,
        parent: StructuralSolver | None = None,
        **_: Any,
    ):
        self.name = name
        self.parent = parent

        self.weight = weight

        self.x_cg = x_cg
        self.y_cg = y_cg
        self.x_cr = x_cr if x_cr is not None else x_cg
        self.y_cr = y_cr if y_cr is not None else y_cg

        self.x_cr_init = self.x_cr
        self.y_cr_init = self.y_cr

        self.draft = 0.0
        self.trim = 0.0
        self.initial_draft = initial_draft
        self.initial_trim = initial_trim

        self._free_dof = np.array([free_in_draft, free_in_trim])
        self._max_disp = np.array([max_draft_step, max_trim_step])
        self._relax = np.array([relax_draft, relax_trim])

        self.loads = GlobalLoads()

        self._res_l = 1.0
        self._res_m = 1.0

        self._motion_solver = RigidBodyMotionSolver(self)

        self._flexible_substructure_residual = 0.0
        self.substructures: list[Substructure] = []

    @cached_property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        if self.parent is None:
            return Config()
        return self.parent.config

    @property
    def ramp(self) -> float:
        """The ramping coefficient from the high-level simulation object."""
        if self.parent is None:
            return 1.0
        return self.parent.simulation.ramp

    @property
    def residual(self) -> float:
        """The combined max residual of lift, moment, and structural node displacement."""
        res_l = self._res_l if self.free_in_draft else 0.0
        res_m = self._res_m if self.free_in_trim else 0.0
        res_node_disp = self._flexible_substructure_residual
        res_torsion = max(
            [
                ss.residual
                for ss in self.substructures
                if isinstance(ss, TorsionalSpringSubstructure)
            ],
            default=0.0,
        )
        return max(res_l, res_m, res_node_disp, res_torsion)

    @property
    def has_free_dof(self) -> bool:
        return self._free_dof.any()

    @property
    def free_in_draft(self) -> bool:
        return self._free_dof[0]

    @free_in_draft.setter
    def free_in_draft(self, value: bool) -> None:
        self._free_dof[0] = value

    @property
    def free_in_trim(self) -> bool:
        return self._free_dof[1]

    @free_in_trim.setter
    def free_in_trim(self, value: bool) -> None:
        self._free_dof[1] = value

    @property
    def motion_file_path(self) -> Path:
        """A path to the file containing the rigid body motinon at a given iteration."""
        return self.parent.simulation.it_dir / f"motion_{self.name}.{self.config.io.data_format}"

    @cached_property
    def nodes(self) -> list[fe.Node]:
        """A list of all unique `Node`s from all component substructures."""
        nodes = set()
        for ss in self.substructures:
            for nd in ss.nodes:
                nodes.add(nd)
        return list(nodes)

    def add_substructure(self, ss: Substructure) -> Substructure:
        """Add a substructure to the rigid body."""
        self.substructures.append(ss)
        ss.rigid_body = self
        return ss

    def get_substructure_by_name(self, name: str) -> Substructure:
        for ss in self.substructures:
            if ss.name == name:
                return ss
        raise LookupError(f"Cannot find substructure with name '{name}'")

    def initialize_position(self) -> None:
        """Initialize the position of the rigid body."""
        self.update_position(self.initial_draft - self.draft, self.initial_trim - self.trim)

    def update_position(self, draft_delta: float = None, trim_delta: float = None) -> None:
        """Update the position of the rigid body by passing the change in draft and trim."""
        if not self.has_free_dof:
            return

        if draft_delta is None or trim_delta is None:
            draft_delta, trim_delta = self.get_disp()
            # TODO: Consider removing this and fixing static types
            assert draft_delta is not None
            assert trim_delta is not None
            if np.isnan(draft_delta):
                draft_delta = 0.0
            if np.isnan(trim_delta):
                trim_delta = 0.0

        for nd in self.nodes:
            coords = nd.coordinates
            new_pos = trig.rotate_point(coords, np.array([self.x_cr, self.y_cr]), trim_delta)
            nd.move(*(new_pos - coords - np.array([0.0, draft_delta])))

        for s in self.substructures:
            s.update_geometry()

        self.x_cg, self.y_cg = trig.rotate_point(
            np.array([self.x_cg, self.y_cg]), np.array([self.x_cr, self.y_cr]), trim_delta
        )
        self.y_cg -= draft_delta
        self.y_cr -= draft_delta

        self.draft += draft_delta
        self.trim += trim_delta

        self.print_motion()

    def _update_flexible_substructure_positions(self) -> None:
        """Update the nodal positions of all component flexible substructures."""
        flexible_substructures = [
            ss for ss in self.substructures if isinstance(ss, FlexibleMembraneSubstructure)
        ]

        num_dof = len(self.parent.nodes) * NUM_DIM
        Kg = np.zeros((num_dof, num_dof))
        Fg = np.zeros((num_dof, 1))
        Ug = np.zeros((num_dof, 1))

        # Assemble global matrices for all substructures together
        for ss in flexible_substructures:
            ss.update_fluid_forces()
            ss.assemble_global_stiffness_and_force(Kg, Fg)

        for nd in self.parent.nodes:
            node_dof = self.parent.node_dofs[nd]
            Fg[node_dof] += nd.fixed_load[:, np.newaxis]

        # Determine fixed degrees of freedom
        dof = [False for _ in Fg]

        for nd in self.parent.nodes:
            node_dof = self.parent.node_dofs[nd]
            for dofi, fdofi in zip(node_dof, nd.is_dof_fixed):
                dof[dofi] = not fdofi

        # Solve FEM linear matrix equation
        if any(dof):
            Ug[np.ix_(dof)] = np.linalg.solve(Kg[np.ix_(dof, dof)], Fg[np.ix_(dof)])

        self._flexible_substructure_residual = np.max(np.abs(Ug))

        Ug *= self.config.solver.relax_FEM
        Ug *= np.min([self.config.solver.max_FEM_disp / np.max(Ug), 1.0])

        for nd in self.parent.nodes:
            node_dof = self.parent.node_dofs[nd]
            nd.move(Ug[node_dof[0], 0], Ug[node_dof[1], 0])

        for ss in flexible_substructures:
            ss.update_geometry()

    def update_substructure_positions(self) -> None:
        """Update the positions of all substructures."""
        self._update_flexible_substructure_positions()
        for ss in self.substructures:
            logger.info(f"Updating position for substructure: {ss.name}")
            if isinstance(ss, TorsionalSpringSubstructure):
                ss.update_angle()

    def update_fluid_forces(self) -> None:
        """Update the fluid forces by summing the force from each substructure."""
        self.loads = GlobalLoads()
        if self.config.body.cushion_force_method.lower() == "assumed":
            self.loads.L += (
                self.config.body.Pc
                * self.config.body.reference_length
                * trig.cosd(self.config.body.initial_trim)
            )

        for ss in self.substructures:
            ss.update_fluid_forces()
            self.loads += ss.loads

        self._res_l = self.get_res_lift()
        self._res_m = self.get_res_moment()

    def get_disp(self) -> np.ndarray:
        """Get the rigid body displacement using Broyden's method."""
        return self._motion_solver.get_disp()

    def get_res_lift(self) -> float:
        """Get the residual of the vertical force balance."""
        if np.isnan(self.loads.L):
            res = 1.0
        else:
            res = (self.loads.L - self.weight) / (
                self.config.flow.stagnation_pressure * self.config.body.reference_length + 1e-6
            )
        return np.abs(res * self.free_in_draft)

    def get_res_moment(self) -> float:
        """Get the residual of the trim moment balance."""
        if np.isnan(self.loads.M):
            res = 1.0
        else:
            if self.x_cg == self.x_cr and self.loads.M == 0.0:
                res = 1.0
            else:
                res = (self.loads.M - self.weight * (self.x_cg - self.x_cr)) / (
                    self.config.flow.stagnation_pressure * self.config.body.reference_length**2
                    + 1e-6
                )
        return np.abs(res * self.free_in_trim)

    def print_motion(self) -> None:
        """Print the moment for debugging."""
        lines = [
            f"Rigid Body Motion: {self.name}",
            f"  CofR: ({self.x_cr}, {self.y_cr})",
            f"  CofG: ({self.x_cg}, {self.y_cg})",
            f"  Draft:      {self.draft:5.4e}",
            f"  Trim Angle: {self.trim:5.4e}",
            f"  Lift Force: {self.loads.L:5.4e}",
            f"  Drag Force: {self.loads.D:5.4e}",
            f"  Moment:     {self.loads.M:5.4e}",
            f"  Lift Force Air: {self.loads.La:5.4e}",
            f"  Drag Force Air: {self.loads.Da:5.4e}",
            f"  Moment Air:     {self.loads.Ma:5.4e}",
            f"  Lift Res:   {self._res_l:5.4e}",
            f"  Moment Res: {self._res_m:5.4e}",
        ]
        for line in lines:
            logger.info(line)

    def write_motion(self) -> None:
        """Write the motion results to file."""
        if self.parent is None:
            raise AttributeError("parent must be set before simulation can be accessed.")
        writers.write_as_dict(
            self.motion_file_path,
            ["xCofR", self.x_cr],
            ["yCofR", self.y_cr],
            ["xCofG", self.x_cg],
            ["yCofG", self.y_cg],
            ["draft", self.draft],
            ["trim", self.trim],
            ["liftRes", self._res_l],
            ["momentRes", self._res_m],
            ["Lift", self.loads.L],
            ["Drag", self.loads.D],
            ["Moment", self.loads.M],
            ["LiftAir", self.loads.La],
            ["DragAir", self.loads.Da],
            ["MomentAir", self.loads.Ma],
        )
        for ss in self.substructures:
            if isinstance(ss, TorsionalSpringSubstructure):
                ss.write_deformation()

    def load_motion(self) -> None:
        """Load the body motion from a file for the current iteration."""
        if self.parent is None:
            raise AttributeError("parent must be set before simulation can be accessed.")
        dict_ = load_dict_from_file(self.motion_file_path)
        self.x_cr = dict_.get("xCofR", np.nan)
        self.y_cr = dict_.get("yCofR", np.nan)
        self.x_cg = dict_.get("xCofG", np.nan)
        self.y_cg = dict_.get("yCofG", np.nan)
        self.draft = dict_.get("draft", np.nan)
        self.trim = dict_.get("trim", np.nan)
        self._res_l = dict_.get("liftRes", np.nan)
        self._res_m = dict_.get("momentRes", np.nan)
        self.loads.L = dict_.get("Lift", np.nan)
        self.loads.D = dict_.get("Drag", np.nan)
        self.loads.M = dict_.get("Moment", np.nan)
        self.loads.La = dict_.get("LiftAir", np.nan)
        self.loads.Da = dict_.get("DragAir", np.nan)
        self.loads.Ma = dict_.get("MomentAir", np.nan)


class RigidBodyMotionSolver:
    """A customized Broyden-method solver to allow manual stepping required for rigid body motion solve."""

    def __init__(self, parent: RigidBody):
        self.parent = parent

        self._jacobian_reset_interval = 6
        self.solver: solver.RootFinder | None = None
        self.disp_old: np.ndarray | None = None
        self.res_old: np.ndarray | None = None
        self.J: np.ndarray | None = None
        self._J_tmp: np.ndarray | None = None
        self._J_fo: np.ndarray | None = None
        self._J_it = 0
        self._step = 0

    def _reset_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Reset the solver Jacobian and modify displacement."""
        f = self._get_residual(x)
        if self._J_tmp is None:
            self._J_tmp = np.zeros((NUM_DIM, NUM_DIM))
            self._J_it = 0
            self._step = 0
            self._J_fo = f
            self.res_old = self._J_fo.copy()
        else:
            self._J_tmp[:, self._J_it] = (f - self._J_fo) / self.disp_old[self._J_it]
            self._J_it += 1

        disp = np.zeros((NUM_DIM,))
        if self._J_it < NUM_DIM:
            disp[self._J_it] = self.parent.config.body.motion_jacobian_first_step

        self.disp_old = disp
        if self._J_it >= NUM_DIM:
            self.J = self._J_tmp.copy()
            self._J_tmp = None
            self.disp_old = None

        return disp

    def _limit_disp(self, disp: np.ndarray) -> np.ndarray:
        """Limit the body displacement.

        If both degrees of freedom are free, we ensure the same ratio of draft-to-trim is maintained.
        If only one is free, or we have a zero displacement in either degree of freedom, then we only step
        in the single direction.

        """
        limited_disp = np.min(np.abs(np.vstack((disp, self.parent._max_disp))), axis=0)

        ind = disp != 0 & self.parent._free_dof
        limit_fraction = min(limited_disp[ind] / np.abs(disp[ind]), default=0.0)

        return disp * limit_fraction * self.parent._free_dof

    def _get_residual(self, _):
        return np.array([self.parent.get_res_lift(), self.parent.get_res_moment()])

    def get_disp(self) -> np.ndarray:
        """Get the rigid body displacement using Broyden's method."""
        x = np.array([self.parent.draft, self.parent.trim])
        if self.solver is None:
            self.solver = RootFinder(self._get_residual, x, method="Broyden")

        if self.J is None:
            return self._reset_jacobian(x)

        if self._step >= self._jacobian_reset_interval:
            logger.debug("Resetting Jacobian for Motion")
            self._reset_jacobian(x)

        f = self._get_residual(x)
        if self.disp_old is not None:
            x += self.disp_old

            dx = self.disp_old[:, np.newaxis]
            df = (f - self.res_old)[:, np.newaxis]
            self.J += (df - self.J @ dx) @ dx.T / np.linalg.norm(dx) ** 2

        dof = self.parent._free_dof
        dx = np.zeros_like(x)
        dx[np.ix_(dof)] = np.linalg.solve(-self.J[np.ix_(dof, dof)], f[np.ix_(dof)])

        disp = self._limit_disp(dx[:] * self.parent._relax)

        self.disp_old = disp
        self.res_old = f[:]
        self._step += 1

        return disp



"""General math helpers."""
from __future__ import annotations

from collections.abc import Callable

import numpy


def sign(x: float) -> float:
    """Return the sign of the argument. Zero returns zero."""
    if x > 0:
        return 1.0
    elif x < 0:
        return -1.0
    else:
        return 0.0


def heaviside(x: float) -> float:
    """The Heaviside step function.

    Returns one if argument is positive, zero if negative, and 0.5 if 0.

    """
    if x > 0.0:
        return 1.0
    elif x < 0.0:
        return 0.0
    else:
        return 0.5


def integrate(x: numpy.ndarray, f: numpy.ndarray) -> float:
    """Integrate a function using Trapezoidal integration."""
    ind = numpy.argsort(x)
    x = x[ind]
    f = f[ind]
    f[numpy.nonzero(numpy.abs(f) == float("Inf"))] = 0.0

    return 0.5 * numpy.sum((x[1:] - x[:-1]) * (f[1:] + f[:-1]))


def deriv(f: Callable[[float], float], x: float, direction: str = "c") -> float:
    """Calculate the derivative of a function at a specific point."""
    dx = 1e-6
    fr = f(x + dx)
    fl = f(x - dx)

    if direction[0].lower() == "r" or numpy.isnan(fl):
        return (fr - f(x)) / dx
    elif direction[0].lower() == "l" or numpy.isnan(fr):
        return (f(x) - fl) / dx
    else:
        return (f(x + dx) - f(x - dx)) / (2 * dx)


def cross2(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Calculate the cross product of two two-dimensional vectors."""
    return a[0] * b[1] - a[1] * b[0]


def cumdiff(x: numpy.ndarray) -> float:
    """Calculate the cumulative difference of an array."""
    return float(numpy.sum(numpy.diff(x)))



from __future__ import annotations

import abc
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Literal
from typing import Type

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import fmin

from planingfsi import logger
from planingfsi import math_helpers
from planingfsi import trig
from planingfsi import writers
from planingfsi.config import NUM_DIM
from planingfsi.fe import felib as fe
from planingfsi.fe.femesh import Subcomponent
from planingfsi.solver import fzero

if TYPE_CHECKING:
    from planingfsi.config import Config
    from planingfsi.fe.rigid_body import RigidBody
    from planingfsi.fe.structure import StructuralSolver
    from planingfsi.potentialflow.pressurepatch import PlaningSurface
    from planingfsi.potentialflow.pressurepatch import PressureCushion


ElementType = ClassVar[Type[fe.Element]]


@dataclass
class GlobalLoads:
    """The integrated global loads on the substructure.

    Attributes:
        D: The hydrodynamic drag.
        L: The hydrodynamic lift.
        M: The hydrodynamic moment (about the body center of rotation).
        Da: The air drag.
        La: The air lift.
        Ma: The air moment (about the body center of rotation).
        Dtip: The drag externally applied to the torsional substructure tip.
        Ltip: The lift externally applied to the torsional substructure tip.
        Mtip: The moment externally applied to the torsional substructure tip
            (about the body center of rotation).

    """

    D: float = 0.0
    L: float = 0.0
    M: float = 0.0
    Da: float = 0.0
    La: float = 0.0
    Ma: float = 0.0
    Dtip: float = 0.0
    Ltip: float = 0.0
    Mtip: float = 0.0

    def __add__(self, other: GlobalLoads) -> GlobalLoads:
        if not isinstance(other, GlobalLoads):
            raise TypeError("Must add GlobalLoads to GlobalLoads")

        return GlobalLoads(
            D=self.D + other.D,
            L=self.L + other.L,
            M=self.M + other.M,
            Da=self.Da + other.Da,
            La=self.La + other.La,
            Ma=self.Ma + other.Ma,
            Dtip=self.Dtip + other.Dtip,
            Ltip=self.Ltip + other.Ltip,
            Mtip=self.Mtip + other.Mtip,
        )


class Substructure(abc.ABC):
    """Base class for a Substructure, which is a component of a RigidBody.

    Attributes:
        name: A name for the substructure.
        seal_pressure: The internal pressure (steady).
        seal_pressure_method: Determines whether to apply a constant or linear hydrostatic internal pressure.
        seal_over_pressure_pct: A factor to apply to the seal_pressure to get the total seal pressure.
        cushion_pressure: If set, this will be the constant external pressure (i.e. for a wetdeck).
        cushion_pressure_type: If there is not interpolator and `cushion_pressure_type` is "Total", then
            a constant external pressure will be applied to the substructure.
        struct_interp_type: The method to use for interpolation of position.
        struct_extrap: If True, extrapolate the position when the arclength extends past the end nodes.
        rigid_body: A reference to the parent `RigidBody` instance.

    """

    is_free = False
    _element_type: ElementType

    def __new__(cls, *, type: str | None = None, **kwargs):
        """This is a factory pattern. If the type is provided, that class will be used
        to create the new object. If not provided, whichever class is instantiated will be used.
        """
        if type is None:
            ss_class = cls
        elif type.lower() == "flexible" or type.lower() == "truss":
            ss_class = FlexibleMembraneSubstructure
        elif type.lower() == "torsionalspring":
            ss_class = TorsionalSpringSubstructure
        else:
            ss_class = RigidSubstructure
        return super().__new__(ss_class)

    def __init__(
        self,
        *,
        name: str = "",
        seal_pressure: float = 0.0,
        seal_pressure_method: Literal["constant"] | Literal["hydrostatic"] = "constant",
        seal_over_pressure_pct: float = 1.0,
        cushion_pressure_type: Literal["Total"] | None = None,
        struct_interp_type: Literal["linear"] | Literal["quadratic"] | Literal["cubic"] = "linear",
        struct_extrap: bool = True,
        solver: StructuralSolver | None = None,
        parent: RigidBody | None = None,
        **_: Any,
    ):
        self.name = name

        self.seal_pressure = seal_pressure
        self.seal_pressure_method = seal_pressure_method
        self.seal_over_pressure_pct = seal_over_pressure_pct

        self.cushion_pressure: float | None = None
        self.cushion_pressure_type = cushion_pressure_type

        self.struct_interp_type = struct_interp_type
        self.struct_extrap = struct_extrap

        self._solver = solver
        self.rigid_body = parent

        # Arrays to store air and fluid pressure profiles for plotting
        self.s_hydro: np.ndarray | None = None
        self.p_hydro: np.ndarray | None = None
        self.s_air: np.ndarray | None = None
        self.p_air: np.ndarray | None = None

        self.elements: list[fe.Element] = []
        self.node_arc_length: np.ndarray | None = None

        self.loads = GlobalLoads()

        self._interpolator: Interpolator | None = None
        self._interp_coords_at_arclength: interp1d | None = None

    @property
    def solver(self) -> StructuralSolver:
        """A reference to the structural solver. Can be explicitly set, or else traverses the parents."""
        if (
            self._solver is None
            and self.rigid_body is not None
            and self.rigid_body.parent is not None
        ):
            return self.rigid_body.parent
        if self._solver is None:
            raise AttributeError("solver must be set before use.")
        return self._solver

    @solver.setter
    def solver(self, solver: StructuralSolver) -> None:
        self._solver = solver

    @property
    def config(self) -> Config:
        """A reference to the simulation configuration."""
        return self.solver.config

    @property
    def ramp(self) -> float:
        """The ramping coefficient from the high-level simulation object."""
        if self.rigid_body is None:
            logger.warning("No parent assigned, ramp will be set to 1.0.")
            return 1.0
        return self.solver.simulation.ramp

    @property
    def nodes(self) -> list[fe.Node]:
        """A list of all nodes in the substructure."""
        nodes = [el.start_node for el in self.elements]
        return nodes + [self.elements[-1].end_node]

    @property
    def arc_length(self) -> float:
        """The total arc length of the substructure, i.e. the sum of all of the Element lengths."""
        return self.node_arc_length[-1]

    @property
    def it_dir(self) -> Path:
        """A path to the current numbered iteration directory, in which to save results."""
        return self.solver.simulation.it_dir

    @property
    def coordinates_file_path(self) -> Path:
        """A path to the coordinates file for the current iteration."""
        return self.it_dir / f"coords_{self.name}.{self.config.io.data_format}"

    def add_planing_surface(self, planing_surface: PlaningSurface, **kwargs: Any) -> PlaningSurface:
        """Add a planing surface to the substructure, and configure the interpolator.

        Args:
            planing_surface: The planing surface.
            **kwargs: Keyword arguments to pass through to the Interpolator.

        Returns:
            The same planing surface passed in as an argument (useful for chaining).

        """
        # Assign the same interpolator to both the substructure and planing_surface
        planing_surface.interpolator = self._interpolator = Interpolator(
            self, planing_surface, **kwargs
        )
        self.solver.simulation.fluid_solver.add_planing_surface(planing_surface)
        return planing_surface

    def add_pressure_cushion(self, pressure_cushion: PressureCushion) -> PressureCushion:
        """Add a constant pressure cushion to the substructure. The cushion acts as a constant
        external pressure applied to the surface, which is integrated into a normal force.

        Args:
            pressure_cushion: The planing surface.

        Returns:
            The same pressure cushion passed in as an argument (useful for chaining).

        """
        self.solver.simulation.fluid_solver.add_pressure_cushion(pressure_cushion)
        self.cushion_pressure_type = "Total"
        self.cushion_pressure = pressure_cushion.cushion_pressure
        return pressure_cushion

    def load_mesh(self, submesh: Path | Subcomponent = Path("mesh")) -> None:
        """Load the mesh from an object or file, and create all component Elements."""
        if isinstance(submesh, Subcomponent):
            nd_idx = [
                (line.start_point.index, line.end_point.index) for line in submesh.line_segments
            ]
        else:
            nd_idx = np.loadtxt(submesh / f"elements_{self.name}.txt", dtype=int)

        self.elements = [
            self._element_type(self.solver.nodes[nd_st_i], self.solver.nodes[nd_end_i], parent=self)
            for nd_st_i, nd_end_i in nd_idx
        ]
        self.update_geometry()

        # Set the interpolation method if there are not enough elements (cubic requires at least 3 elements)
        # else keep what was specified in the constructor
        if len(self.elements) == 1:
            self.struct_interp_type = "linear"
        elif len(self.elements) == 2 and not self.struct_interp_type == "linear":
            self.struct_interp_type = "quadratic"

    def update_geometry(self) -> None:
        """Update geometry and interpolation functions in the process."""
        element_lengths = [el.length for el in self.elements]
        self.node_arc_length = np.cumsum([0.0] + element_lengths)

        nodal_coordinates = np.array([nd.coordinates for nd in self.nodes])
        self._interp_coords_at_arclength = interp1d(
            self.node_arc_length,
            nodal_coordinates.T,
            kind=self.struct_interp_type,
            fill_value="extrapolate" if self.struct_extrap else np.nan,
        )

    def get_coordinates(self, si: float) -> np.ndarray:
        """Return the coordinates of the surface at a specific arclength."""
        return self._interp_coords_at_arclength(si)

    def write_coordinates(self) -> None:
        """Write the coordinates of all component nodes to file."""
        writers.write_as_list(
            self.coordinates_file_path,
            ["x [m]", [nd.x for nd in self.nodes]],
            ["y [m]", [nd.y for nd in self.nodes]],
        )

    def load_coordinates(self) -> None:
        """Set the coordinates of each node by loading from the saved file from the current iteration."""
        coordinates = np.loadtxt(self.coordinates_file_path)
        for coords, nd in zip(coordinates, self.nodes):
            nd.coordinates = coords
        self.update_geometry()

    def _get_loads_in_range(self, s_start, s_end, /):
        if self._interpolator is not None:
            return self._interpolator.get_loads_in_range(
                s_start,
                s_end,
                pressure_limit=(
                    self.config.flow.stagnation_pressure
                    if self.config.plotting.pressure_limiter
                    else None
                ),
            )
        else:
            s = np.array([s_start, s_end])
            return s, np.zeros_like(s), np.zeros_like(s)

    def update_fluid_forces(self) -> None:
        """Update the fluid forces on all structural elements.

        The pressure is interpolated, both hydrostatic and air, and applied in various
        manners depending on the type of substructure.

        """

        s_hydro: list[float] = []
        p_hydro: list[float] = []
        s_air: list[float] = []
        p_air: list[float] = []
        self.loads = GlobalLoads()
        if self._interpolator is not None:
            s_min, s_max = self._interpolator.get_min_max_s()

        for i, el in enumerate(self.elements):
            # Get pressure & shear stress at end points and all fluid points along element
            s_start, s_end = self.node_arc_length[i], self.node_arc_length[i + 1]
            s, pressure_hydro, tau = self._get_loads_in_range(s_start, s_end)

            # Apply ramp to hydrodynamic pressure
            pressure_hydro *= self.ramp**2

            # Add external cushion pressure to external fluid pressure
            # This is the full-resolution calculation with all structural nodes and fluid elements
            pressure_cushion = np.zeros_like(s)
            if self._interpolator is not None:
                pressure_cushion[s > s_max] = self._interpolator.fluid.upstream_pressure
                pressure_cushion[s < s_min] = self._interpolator.fluid.downstream_pressure
            elif self.cushion_pressure_type == "Total":
                pressure_cushion[:] = self.cushion_pressure or self.config.body.Pc

            # Calculate internal pressure
            pressure_internal = np.full_like(s, self.seal_pressure * self.seal_over_pressure_pct)
            if self.seal_pressure_method.lower() == "hydrostatic":
                pressure_internal += (
                    self.config.flow.density
                    * self.config.flow.gravity
                    * (
                        self.config.flow.waterline_height
                        - np.array([self.get_coordinates(si)[1] for si in s])
                    )
                )

            # Derive various combinations of pressure
            pressure_air_net = pressure_internal - pressure_cushion
            pressure_external = pressure_hydro + pressure_cushion
            pressure_total = pressure_external - pressure_internal

            # Store fluid and air pressure components for element (for plotting)
            if i == 0:
                s_hydro.append(s[0])
                p_hydro.append(pressure_hydro[0])
                s_air.append(s[0])
                p_air.append(pressure_air_net[0])

            s_hydro.extend(s[1:])
            p_hydro.extend(pressure_hydro[1:])
            s_air.append(s[-1])
            p_air.append(pressure_air_net[-1])

            if isinstance(self, FlexibleMembraneSubstructure):
                el.qp = self._distribute_integrated_load_to_nodes(s, pressure_total)
                el.qs = self._distribute_integrated_load_to_nodes(s, -tau)

            # Calculate external force and moment for rigid body calculation
            if self._interpolator is not None:
                self.loads.D = self._interpolator.fluid.drag_total
                self.loads.L = self._interpolator.fluid.lift_total
                self.loads.M = self._interpolator.fluid.moment_total
            else:
                method_pressure_map = {"integrated": pressure_external, "assumed": pressure_hydro}
                if (
                    p := method_pressure_map.get(self.config.body.cushion_force_method.lower())
                ) is not None:
                    f_x, f_y, moment = self._get_integrated_global_loads(s, p, tau)
                    self.loads.D -= f_x
                    self.loads.L += f_y
                    self.loads.M += moment

            # Integrate the total pressure for torsional spring calculations
            if isinstance(self, TorsionalSpringSubstructure):
                _, _, moment = self._get_integrated_global_loads(
                    s, pressure_total, tau, moment_about=self.base_pt
                )
                self._applied_moment += moment

            # Integrate global cushion pressure force and moment
            f_x, f_y, moment = self._get_integrated_global_loads(s, pressure_cushion)
            self.loads.Da -= f_x
            self.loads.La += f_y
            self.loads.Ma += moment

        self.p_hydro = np.array(p_hydro)
        self.s_hydro = np.array(s_hydro)
        self.p_air = np.array(p_air)
        self.s_air = np.array(s_air)

    @staticmethod
    def _distribute_integrated_load_to_nodes(s: np.ndarray, load: np.ndarray) -> np.ndarray:
        """Integrate a load along an element, returning the equivalent load at each endpoint."""
        integral = math_helpers.integrate(s, load)
        if integral == 0.0:
            return np.zeros(2)
        pct = (math_helpers.integrate(s, s * load) / integral - s[0]) / math_helpers.cumdiff(s)
        return integral * np.array([1 - pct, pct])

    def _get_integrated_global_loads(
        self,
        s: np.ndarray,
        p: np.ndarray,
        tau: np.ndarray | None = None,
        *,
        moment_about: np.ndarray | None = None,
    ) -> tuple[float, float, float]:
        """Integrate a pressure and shear stress along an arclength to get global loads.

        The moment is calculated about the center of rotation unless an alternative point is provided.

        Args:
            s: The arclength array.
            p: The pressure array.
            tau: The shear stress array.
            moment_about: The point about which to calculate the moment.

        Returns:
            A tuple containing x-force, y-force, and z-moment.

        """
        if tau is None:
            tau = np.zeros_like(p)

        n = np.array([self.get_normal_vector(s_i) for s_i in s])
        t = np.array([trig.rotate_vec_2d(n_i, -90) for n_i in n])
        f = -p[:, np.newaxis] * n + tau[:, np.newaxis] * t

        assert self.rigid_body is not None
        if moment_about is None:
            moment_about = np.array([self.rigid_body.x_cr, self.rigid_body.y_cr])

        r = [self.get_coordinates(s_i) - moment_about for s_i in s]
        m = np.array([math_helpers.cross2(r_i, f_i) for r_i, f_i in zip(r, f)])

        drag = math_helpers.integrate(s, f[:, 0])
        lift = math_helpers.integrate(s, f[:, 1])
        moment = math_helpers.integrate(s, m)
        return drag, lift, moment

    def get_normal_vector(self, s: float) -> np.ndarray:
        """Calculate the normal vector at a specific arc length."""
        dx_ds = math_helpers.deriv(lambda si: self.get_coordinates(si)[0], s)
        dy_ds = math_helpers.deriv(lambda si: self.get_coordinates(si)[1], s)

        return trig.rotate_vec_2d(trig.angd2vec2d(trig.atand2(dy_ds, dx_ds)), -90)

    def fix_all_degrees_of_freedom(self) -> None:
        """Set all degrees of freedom of all nodes in the substructure."""
        for nd in self.nodes:
            nd.is_dof_fixed = tuple(True for _ in range(NUM_DIM))


class RigidSubstructure(Substructure):
    """A substructure that is rigidly attached to the rigid body.

    This means that the substructure moves with the rigid body but does not deform.

    """

    _element_type: ElementType = fe.RigidElement

    def load_mesh(self, submesh: Path | Subcomponent = Path("mesh")) -> None:
        super().load_mesh(submesh)
        self.fix_all_degrees_of_freedom()


class FlexibleMembraneSubstructure(Substructure):
    """A flexible membrane structure, composed of truss finite elements.

    The structure supports axial loading only (i.e. no bending), and supports large deformations.

    Attributes:
        pretension: An optional pretension used to initialize the elements.
        EA: The axial stiffness of the component `Element`s.

    """

    is_free = True
    _element_type: ElementType = fe.TrussElement
    elements: list[fe.TrussElement]

    def __init__(
        self,
        *,
        pretension: float = -0.5,
        axial_stiffness: float = 5e7,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.pretension = pretension
        self.EA = axial_stiffness

    def assemble_global_stiffness_and_force(
        self, K_global: np.ndarray, F_global: np.ndarray
    ) -> None:
        """Assemble element stiffness and force into the global matrices, which are passed by reference and modified."""
        for el in self.elements:
            K_el, F_el = el.get_stiffness_and_force()
            el_dof = [dof for nd in el.nodes for dof in self.rigid_body.parent.node_dofs[nd]]
            K_global[np.ix_(el_dof, el_dof)] += K_el
            F_global[np.ix_(el_dof)] += F_el

    def load_mesh(self, submesh: Path | Subcomponent = Path("mesh")) -> None:
        """Load the mesh, and assign structural properties to the elements."""
        super().load_mesh(submesh)
        for el in self.elements:
            el.initial_axial_force = -self.pretension
            el.EA = self.EA


class TorsionalSpringSubstructure(Substructure):
    """A substructure that is locally rigid but is attached to the rigid body with a torsional spring.

    Thus, the nodes can move, but rotate about a base point with a linear spring stiffness.

    Attributes:
        initial_angle: The initial offset angle of the structure, where 0 is the equilibrium angle
            with no load applied.
        tip_load: The vertical load to apply at a specified fractional arclength.
        tip_load_pct: The fraction along arclength at which to apply a fixed tip load.
        base_pt_pct: The fraction along arclength around which the substructure rotates.
        spring_constant: The linear spring constant [(N.m)/deg].
        relaxation_angle: A relaxation factor to apply to the angle change.
        minimum_angle: The minimum allowable angle for the substructure.
        max_angle_step: The maximum absolute change in angle during a single iteration.

    """

    base_pt: np.ndarray
    is_free = True
    _element_type: ElementType = fe.RigidElement

    def __init__(
        self,
        *,
        initial_angle: float = 0.0,
        tip_load: float = 0.0,
        tip_load_pct: float = 0.0,
        base_pt_pct: float = 1.0,
        spring_constant: float = 1e3,
        relaxation_angle: float | None = None,
        minimum_angle: float = -float("Inf"),
        max_angle_step: float = float("Inf"),
        attach_pct: float = 0.0,
        attached_substructure_name: str | None = None,
        attached_substructure_end: Literal["start"] | Literal["end"] = "end",
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.tip_load = tip_load
        self.tip_load_pct = tip_load_pct
        self.base_pt_pct = base_pt_pct
        if spring_constant <= 0.0:
            raise ValueError("Spring constant must be positive.")
        self.spring_constant = spring_constant

        self._theta = initial_angle
        self._relax = relaxation_angle
        self.attach_pct = attach_pct
        self.attached_node: fe.Node | None = None
        self.attached_element: fe.Element | None = None
        self.minimum_angle = minimum_angle
        self.max_angle_step = max_angle_step
        self.attached_substructure_name = attached_substructure_name
        self.attached_substructure_end = attached_substructure_end
        self.attached_ind = 0
        self.attached_substructure: Substructure | None = None
        self.residual = 1.0
        self._applied_moment = 0.0

    @property
    def relax(self) -> float:
        """Relaxation parameter."""
        if self._relax is None:
            return self.config.body.relax_rigid_body
        return self._relax

    @property
    def angle(self) -> float:
        """The current rotation angle of the substructure."""
        return self._theta

    @angle.setter
    def angle(self, value: float) -> None:
        nodes = self.nodes.copy()
        if self.attached_node is not None and self.attached_node not in self.nodes:
            nodes.append(self.attached_node)

        angle_change = np.max([value, self.minimum_angle]) - self._theta
        for nd in nodes:
            nd.coordinates = trig.rotate_point(nd.coordinates, self.base_pt, -angle_change)

        self._theta += angle_change
        self.residual = abs(angle_change)
        self.update_geometry()
        logger.info(f"  Deformation for substructure {self.name}: {self._theta}")

    def load_mesh(self, submesh: Path | Subcomponent = Path("mesh")) -> None:
        super().load_mesh(submesh)
        self.fix_all_degrees_of_freedom()
        if self.base_pt_pct == 1.0:
            self.base_pt = self.nodes[-1].coordinates
        elif self.base_pt_pct == 0.0:
            self.base_pt = self.nodes[0].coordinates
        else:
            self.base_pt = self.get_coordinates(self.base_pt_pct * self.arc_length)

        self._set_attachments()

    def _set_attachments(self) -> None:
        if self.attached_substructure_name is not None:
            self.attached_substructure = self.rigid_body.get_substructure_by_name(
                self.attached_substructure_name
            )

        if self.attached_substructure_end == "start":
            self.attached_ind = 0
        else:
            self.attached_ind = -1

        if self.attached_node is None and self.attached_substructure is not None:
            self.attached_node = self.attached_substructure.nodes[self.attached_ind]
            self.attached_element = self.attached_substructure.elements[self.attached_ind]

    def _apply_tip_load(self) -> None:
        """Apply any externally tip load."""
        tip_coords = self.get_coordinates(self.tip_load_pct * self.arc_length)
        tip_rel_coords = tip_coords - self.base_pt
        tip_force = np.array([0.0, self.tip_load]) * self.ramp
        tip_moment = math_helpers.cross2(tip_rel_coords, tip_force)
        self._applied_moment += tip_moment
        # Add to the global forces
        self.loads.Ltip += tip_force[1]
        self.loads.Mtip += math_helpers.cross2(
            tip_coords - np.array([self.rigid_body.x_cr, self.rigid_body.x_cr]), tip_force
        )

    def update_fluid_forces(self) -> None:
        self._applied_moment = 0.0
        super().update_fluid_forces()
        self._apply_tip_load()

    def update_angle(self) -> None:
        """Update the angle of the substructure."""
        if np.isnan(self._applied_moment):
            theta = 0.0
        else:
            theta = -self._applied_moment / self.spring_constant

        # Ramp and limit the angle change
        angle_change = (theta - self._theta) * self.relax
        angle_change = min(abs(angle_change), self.max_angle_step) * np.sign(angle_change)

        self.angle = self._theta + angle_change

    def write_deformation(self) -> None:
        """Write the deformation angle to a file."""
        writers.write_as_dict(
            self.it_dir / f"deformation_{self.name}.{self.config.io.data_format}",
            ["angle", self._theta],
        )


class Interpolator:
    """An object that handles two-way communication between the structural and fluid solvers."""

    def __init__(
        self,
        solid: Substructure,
        fluid: PlaningSurface,
        *,
        waterline_height: float = 0.0,
        separation_arclength_start_pct: float = 0.5,
        immersion_arclength_start_pct: float = 0.9,
        **_: Any,
    ):
        self.solid = solid
        self.fluid = fluid

        self._waterline_height = waterline_height
        self._separation_arclength_start_pct = separation_arclength_start_pct
        self._immersion_arclength_start_pct = immersion_arclength_start_pct

        self._separation_arclength: float | None = None
        self._immersed_arclength: float | None = None

    def get_body_height(self, x: float) -> float:
        """Return the elevation (height) of the substructure surface at a given x-coordinate.

        Args:
            x: The x-coordinate.

        Returns:
            The elevation of the surface.

        """
        s = np.max([self.get_s_fixed_x(x), 0.0])
        return self.get_coordinates(s)[1]

    def get_coordinates(self, s: float) -> np.ndarray:
        """The surface coordinates at a given arclength."""
        return self.solid.get_coordinates(s)

    def get_min_max_s(self) -> tuple[float, float]:
        """Return the min and max arclength that the fluid elements cover.

        Returns:
            A tuple of (min, max) arclength.

        """
        pts = self.fluid.element_coords
        return self.get_s_fixed_x(pts[0]), self.get_s_fixed_x(pts[-1])

    def get_s_fixed_x(self, x: float, so_pct: float = 0.5) -> float:
        """Return the arclength of the substructure surface at a given x-coordinate.

        Args:
            x: The x-coordinate.
            so_pct: The starting arclength guess, as a fraction of the total arclength.

        Returns:
            The arclength of the surface.

        """
        return fzero(lambda s: self.get_coordinates(s)[0] - x, so_pct * self.solid.arc_length)

    @property
    def immersed_length(self) -> float:
        """The total immersed length of the substructure.

        This is the length of the surface below the undisturbed waterline.

        """
        if self._immersed_arclength is None:
            self._immersed_arclength = self._immersion_arclength_start_pct * self.solid.arc_length

        self._immersed_arclength = fzero(
            lambda s: self.get_coordinates(s)[1] - self._waterline_height,
            self._immersed_arclength,
        )

        return self.get_coordinates(self._immersed_arclength)[0]

    def get_separation_point(self) -> np.ndarray:
        """Return the coordinates of the assumed separation point.

        The separation is assumed to occur at the point of lowest elevation along the surface.

        Returns:
            A 2-d array containing the (x,y) coordinates of the separation point.

        """

        def get_y_coords(s: float) -> float:
            return self.get_coordinates(s)[1]

        if self._separation_arclength is None:
            self._separation_arclength = (
                self._separation_arclength_start_pct * self.solid.arc_length
            )

        self._separation_arclength = fmin(
            get_y_coords, self._separation_arclength, disp=False, xtol=1e-6
        )[0]
        self._separation_arclength = float(np.max([self._separation_arclength, 0.0]))
        return self.get_coordinates(self._separation_arclength)

    def get_loads_in_range(
        self, s0: float, s1: float, /, *, pressure_limit: float | None = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get the fluid pressure and shear stress at any pressure elements within the provided arclength range.

        Args:
            s0: The lower-bound arclength.
            s1: The upper-bound arclength.
            pressure_limit: Optionally limit the pressure to be below a specified value (e.g. stagnation pressure).

        Returns:
            A tuple of arrays containing the arclength, pressure, and shear stress at each pressure element point.

        """
        x, p, tau = self.fluid.get_loads_in_range(
            self.solid.get_coordinates(s0)[0],
            self.solid.get_coordinates(s1)[0],
            pressure_limit=pressure_limit,
        )
        s = np.array([self.get_s_fixed_x(xx) for xx in x])
        s[0], s[-1] = s0, s1  # Force ends to be exactly the same
        return s, p, tau



"""Command line program for PlaningFSI.

This script is called from the command line with options to plot or load
results from file. It reads the fluid and solid body properties from dictionary
files, assembles the problem, and runs it.

"""
from __future__ import annotations

from pathlib import Path

import click
import click_log

from planingfsi import logger
from planingfsi.config import Config
from planingfsi.fe.femesh import Mesh
from planingfsi.simulation import Simulation


@click.group(name="planingfsi", help="Run the PlaningFSI program")
def cli() -> None:
    click_log.basic_config(logger)


@cli.command(name="run")
@click.option(
    "post_mode",
    "--post",
    is_flag=True,
    help="Run in post-processing mode, loading results from file and generating figures.",
)
@click.option("--plot-save", is_flag=True, help="Save the plots to figures.")
@click.option("--plot-show", is_flag=True, help="Show the plots in a pop-up window.")
@click.option(
    "new_case",
    "--new",
    is_flag=True,
    help="Force generate new case, deleting old results first.",
)
def run_planingfsi(post_mode: bool, plot_save: bool, plot_show: bool, new_case: bool) -> None:
    """Run the planingFSI solver."""
    simulation = Simulation.from_input_files("configDict")

    simulation.config.plotting.save = plot_save
    simulation.config.plotting.show = plot_show

    if post_mode:
        logger.info("Running in post-processing mode")
        simulation.config.plotting.save = True
        simulation.config.io.results_from_file = True

    if new_case:
        logger.info("Removing all time directories")
        for it_dir in Path(simulation.config.path.case_dir).glob("[0-9]*"):
            it_dir.unlink()

    simulation.run()


@cli.command(name="mesh")
@click.argument("mesh_dict", required=False)
@click.option("plot_show", "--show", is_flag=True)
@click.option("plot_save", "--save", is_flag=True)
@click.option("--verbose", is_flag=True)
def generate_mesh(mesh_dict: str | None, plot_show: bool, plot_save: bool, verbose: bool) -> None:
    """Generate the initial mesh."""
    config = Config.from_file("configDict")

    if mesh_dict is not None:
        config.path.mesh_dict_name = mesh_dict

    if not Path(config.path.mesh_dict_name).exists():
        raise FileNotFoundError(f"File {config.path.mesh_dict_name} does not exist")

    mesh = Mesh()
    with Path(config.path.mesh_dict_name).open() as fp:
        exec(fp.read())

    mesh.display(disp=verbose)
    mesh.plot(show=plot_show, save=plot_save)
    mesh.write(mesh_dir=config.path.mesh_dir_name)



"""Convenient trigonometric functions."""
import math
from typing import Union

import numpy

ArrayLike = Union[float, numpy.ndarray]


def cosd(ang: ArrayLike) -> ArrayLike:
    """Return the cosine of an angle specified in degrees.

    Parameters
    ----------
    ang : float
       An angle, in degrees.

    """
    return numpy.cos(numpy.radians(ang))


def sind(ang: ArrayLike) -> ArrayLike:
    """Return the sine of an angle specified in degrees.

    Parameters
    ----------
    ang : float
       An angle, in degrees.

    """
    return numpy.sin(numpy.radians(ang))


def tand(ang: ArrayLike) -> ArrayLike:
    """Return the tangent of an angle specified in degrees.

    Parameters
    ----------
    ang : float
       An angle, in degrees.

    """
    return math.tan(numpy.radians(ang))


def acosd(slope: ArrayLike) -> ArrayLike:
    """Return the arccosine of an angle specified in degrees.

    Returns
    -------
    float
       An angle, in degrees.

    """
    return numpy.degrees(numpy.arccos(slope))


def asind(slope: ArrayLike) -> ArrayLike:
    """Return the arcsine of an angle specified in degrees.

    Returns
    -------
    float
       An angle, in degrees.

    """
    return numpy.degrees(numpy.arcsin(slope))


def atand(slope: ArrayLike) -> ArrayLike:
    """Return the arctangent of an angle specified in degrees.

    Returns
    -------
    float
       An angle, in degrees.

    """
    return numpy.degrees(numpy.arctan(slope))


def atand2(delta_y: ArrayLike, delta_x: ArrayLike) -> float:
    """Return the arctan2 of an angle specified in degrees.

    Returns
    -------
    float
       An angle, in degrees.

    """
    return float(numpy.degrees(numpy.arctan2(delta_y, delta_x)))


def ang2vec(ang: ArrayLike) -> numpy.ndarray:
    """Convert angle in radians to a 3-d unit vector in the x-y plane."""
    return numpy.array([numpy.cos(ang), numpy.sin(ang), 0.0])


def ang2vecd(ang: float) -> numpy.ndarray:
    """Convert angle in degrees to a 3-d unit vector in the x-y plane."""
    return ang2vec(numpy.radians(ang))


def angd2vec2d(ang: float) -> numpy.ndarray:
    """Convert angle in degrees to a 2-d unit vector in the x-y plane."""
    return ang2vecd(ang)[:2]


def rotate_vec_2d(vec: numpy.ndarray, ang: float) -> numpy.ndarray:
    """Rotate a 2d vector v by angle ang in degrees."""
    vec3d = numpy.zeros(3)
    vec3d[:2] = vec
    return rotate_vec(vec3d, 0, 0, ang)[:2]


def rotate_vec(
    vec: numpy.ndarray,
    ang_x: float = 0.0,
    ang_y: float = 0.0,
    ang_z: float = 0.0,
    about: numpy.ndarray = numpy.zeros(3),
) -> numpy.ndarray:
    """Rotate a 3d vector v by angle ang in degrees."""
    rot_x = numpy.array([[1, 0, 0], [0, cosd(ang_x), -sind(ang_x)], [0, sind(ang_x), cosd(ang_x)]])
    rot_y = numpy.array([[cosd(ang_y), 0, sind(ang_y)], [0, 1, 0], [-sind(ang_y), 0, cosd(ang_y)]])
    rot_z = numpy.array([[cosd(ang_z), -sind(ang_z), 0], [sind(ang_z), cosd(ang_z), 0], [0, 0, 1]])

    return (
        numpy.dot(numpy.dot(numpy.dot(rot_x, rot_y), rot_z), (numpy.asarray(vec) - about).T) + about
    )


def rotate_point(point: numpy.ndarray, about: numpy.ndarray, angle: float) -> numpy.ndarray:
    """Rotate a point about another point by a specific angle in degrees."""
    relative_pos = numpy.array(point) - numpy.array(about)
    new_pos = rotate_vec_2d(relative_pos, angle)
    return about + new_pos



from __future__ import annotations

import abc
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable

import numpy as np

from planingfsi import logger
from planingfsi.config import NUM_DIM

if TYPE_CHECKING:
    from planingfsi.fe.substructure import Substructure


class Node:
    """A finite-element node, used to represent the end coordinates of truss elements.

    Attributes:
        coordinates: An array of (x, y) coordinates representing the nodal location.
        is_dof_fixed: A length-two tuple corresponding to whether the node is fixed in (x, y), respectively.
        fixed_load: An array of (x, y) external forces to apply to the node.

    """

    def __init__(
        self,
        coordinates: np.ndarray,
        *,
        is_dof_fixed: Iterable[bool] | None = None,
        fixed_load: np.ndarray | None = None,
    ) -> None:
        self.coordinates = np.array(coordinates, dtype=np.float64)
        if is_dof_fixed is not None:
            self.is_dof_fixed = tuple(bool(dof) for dof in is_dof_fixed)
        else:
            self.is_dof_fixed = tuple([False] * NUM_DIM)
        self.fixed_load = (
            np.array(fixed_load, dtype=np.float64) if fixed_load is not None else np.zeros(NUM_DIM)
        )

    @property
    def x(self) -> float:
        """The x-coordinate."""
        return self.coordinates[0]

    @property
    def y(self) -> float:
        """The y-coordinate."""
        return self.coordinates[1]

    def move(self, dx: float, dy: float) -> None:
        """Move the node by a given displacement in x & y directions.

        Args:
            dx: The displacement in x-direction.
            dy: The displacement in y-direction.

        """
        self.coordinates += np.array([dx, dy])


class Element(abc.ABC):
    """A generic element.

    Attributes:
        start_node: The start node.
        end_node: The end node.
        initial_length: The initial length of the element, i.e. distance between start & end node.
        qp: The external forces applied in the perpendicular direction.
        qs: The external forces applied in the shear direction.
        parent: A reference to the parent substructure.

    """

    def __init__(self, start_node: Node, end_node: Node, *, parent: Substructure | None = None):
        self.start_node = start_node
        self.end_node = end_node
        self.initial_length = self.length
        self.qp = np.zeros(2)
        self.qs = np.zeros(2)
        self.parent = parent
        self._initial_coordinates = [nd.coordinates for nd in self.nodes]  # save for plotting

    @property
    def nodes(self) -> tuple[Node, Node]:
        """A tuple containing both nodes in the Element."""
        return self.start_node, self.end_node

    @property
    def ramp(self) -> float:
        """The ramping coefficient from the high-level simulation object."""
        if self.parent is None:
            logger.warning("No parent assigned, ramp will be set to 1.0.")
            return 1.0
        return self.parent.ramp

    @property
    def length(self) -> float:
        """The element length, or Cartesian distance between nodes."""
        return np.linalg.norm(self.end_node.coordinates - self.start_node.coordinates)


class RigidElement(Element):
    """A rigid element, which is not subject to deformation and is generally used for drawing."""


class TrussElement(Element):
    """A truss element, used for large-deformation membrane structures.

    Attributes:
        initial_axial_force: The axial force in the element at the beginning of the simulation.
            Used to apply pre-tension. This value is ramped up to avoid numerical instability.
        EA: The axial stiffness of the element.

    """

    def __init__(self, *args, initial_axial_force: float = 0.0, EA: float = 0.0, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.initial_axial_force = initial_axial_force
        self.EA = EA

    @property
    def axial_force(self) -> float:
        """The axial force in the element due to extension/compression."""
        axial_force = (1.0 - self.ramp) * self.initial_axial_force
        axial_force += self.EA * (self.length - self.initial_length) / self.initial_length
        return axial_force

    def get_stiffness_and_force(self) -> tuple[np.ndarray, np.ndarray]:
        """The elemental stiffness matrix and force vector, in the global coordinate system."""
        # Stiffness matrices in local coordinates
        stiffness_linear_local = (
            np.array([[1, 0, -1, 0], [0, 0, 0, 0], [-1, 0, 1, 0], [0, 0, 0, 0]])
            * self.EA
            / self.length
        )

        stiffness_nonlinear_local = (
            np.array([[1, 0, -1, 0], [0, 1, 0, -1], [-1, 0, 1, 0], [0, -1, 0, 1]])
            * self.axial_force
            / self.length
        )

        # Force vectors in local coordinates
        force_linear_local = np.array([[self.qs[0]], [self.qp[0]], [self.qs[1]], [self.qp[1]]])
        force_nonlinear_local = np.array([[1], [0], [-1], [0]]) * self.axial_force

        # Add linear and nonlinear components
        stiffness_total_local = stiffness_linear_local + stiffness_nonlinear_local
        force_total_local = force_linear_local + force_nonlinear_local

        # Rotate stiffness and force matrices into global coordinates
        gamma = np.arctan2(self.end_node.y - self.start_node.y, self.end_node.x - self.start_node.x)
        c, s = np.cos(gamma), np.sin(gamma)

        transformation_matrix = np.array([[c, s, 0, 0], [-s, c, 0, 0], [0, 0, c, s], [0, 0, -s, c]])

        stiffness_total_global = (
            transformation_matrix.T @ stiffness_total_local @ transformation_matrix
        )
        force_total_global = transformation_matrix.T @ force_total_local

        return stiffness_total_global, force_total_global



"""This module is used to store the global configuration classes.

Values are generally read on from the `configDict` file, which is
parsed into separate domain configurations. Each of these domain
configurations are stored under the high-level `Config` instance.
In general, the configuration is attached to the `Simulation`
instance, which then serves as a reference point elsewhere in the code.

"""
from __future__ import annotations

import math
from pathlib import Path
from typing import Any

from planingfsi import logger
from planingfsi.dictionary import load_dict_from_file

DICT_NAME = "configDict"
NUM_DIM = 2


class ConfigItem:
    """A descriptor to represent a configuration item.

    Attributes are loaded from a dictionary with fancy default handling.

    """

    name: str
    default: Any

    def __init__(self, *alt_keys: str, **kwargs: Any):
        self.alt_keys: list[str] = list(alt_keys)
        self.type_: type | None = kwargs.get("type")
        try:
            self.default = kwargs["default"]
            if self.default is not None:
                self.type_ = type(self.default)
        except KeyError:
            # If default not specified, leave undefined. Handled in __get__.
            # Allows None to be the default value.
            pass

    def __get__(self, instance: "SubConfig", _: Any) -> Any:
        """Retrieve the value from the instance dictionary or return the default.

        Raises:
            AttributeError: If the attribute hasn't been set and the  default value isn't specified.

        """
        if self.name in instance.__dict__:
            return instance.__dict__[self.name]
        try:
            return self.default
        except AttributeError:
            raise AttributeError(f"Attribute {self.name} has not been set and no default specified")

    def __set__(self, instance: "SubConfig", value: Any) -> None:
        """When the value is set, try to convert it and then store it in the instance dictionary."""
        if value is not None and self.type_ is not None:
            if self.type_ == bool and isinstance(value, str):
                value = value.lower() == "true"
            value = self.type_(value)
        instance.__dict__[self.name] = value

    def __set_name__(self, _: Any, name: str) -> None:
        """Set the name when the ConfigItem is defined."""
        self.name = name

    @property
    def keys(self) -> list[str]:
        """A list of keys to look for when reading in the value."""
        return [self.name] + self.alt_keys

    def get_from_dict(self, dict_: dict[str, Any]) -> Any:
        """Try to read all keys from the dictionary until a non-None value is found.

        Returns the default value if no appropriate value is found in the dictionary.

        """
        for key in self.keys:
            value = dict_.get(key)
            if value is not None:
                return value
        raise KeyError('None of the following keys "{}" found in dictionary.'.format(self.keys))


class SubConfig:
    """An empty class used simply for dividing the configuration into
    different sections. Also useful in helping define the namespace scopes.
    """

    def __init__(self, parent: Config | None = None):
        self._parent = parent

    @property
    def parent(self) -> "Config":
        """Raises a ValueError if parent is not assigned."""
        if self._parent is None:
            raise ValueError("Must assign a parent to access this property.")
        return self._parent

    def load_from_file(self, filename: Path | str) -> None:
        """Load the configuration from a dictionary file.

        Args:
            filename: The path to the dictionary file.

        """
        # Clear the attributes before loading the new ones
        # TODO: This is hacky
        self.__dict__ = {"_parent": self.parent}
        dict_ = load_dict_from_file(filename)
        for key, config_item in self.__class__.__dict__.items():
            if isinstance(config_item, ConfigItem):
                try:
                    value = config_item.get_from_dict(dict_)
                except KeyError:
                    pass
                else:
                    setattr(self, config_item.name, value)


class FlowConfig(SubConfig):
    """Configuration related to the fluid dynamics problem.

    Attributes:
        density (float): Mass density of the fluid.
        gravity (float): Acceleration due to gravity.
        kinematic_viscosity (float): Kinematic viscosity of the fluid.
        waterline_height (float): Height of the waterline above the reference.
        include_friction (bool): If True, include a flat-plate estimation for the frictional drag
            component.

    """

    density = ConfigItem("rho", default=998.2)
    gravity = ConfigItem("g", default=9.81)
    kinematic_viscosity = ConfigItem("nu", default=1e-6)
    waterline_height = ConfigItem("hWL", default=0.0)
    include_friction = ConfigItem("shearCalc", default=False)

    _froude_num = ConfigItem("Fr", default=None, type=float)
    _flow_speed = ConfigItem("U", default=None, type=float)

    @property
    def reference_length(self) -> float:
        """float: Reference length used for potential-flow solver.

        Defaults to reference length of the rigid body.

        """
        return self.parent.body.reference_length

    @property
    def flow_speed(self) -> float:
        """The flow speed is the native variable to store free-stream velocity. However, if Froude
        number is set from input file, that should override the flow speed input.

        """
        if self._froude_num is not None:
            if self._flow_speed is not None:
                raise ValueError(
                    "Only one flow speed variable (either Froude number or flow "
                    "speed) must be set in {0}".format(DICT_NAME)
                )
            self.froude_num = self._froude_num
        elif self._flow_speed is None:
            raise ValueError("Must specify either U or Fr in {0}".format(DICT_NAME))
        return getattr(self, "_flow_speed")

    @flow_speed.setter
    def flow_speed(self, value: float) -> None:
        """Set the raw flow speed variable and ensure raw Froude number is not also set."""
        self._flow_speed = value
        self._froude_num = None

    @property
    def froude_num(self) -> float:
        """float: The Froude number is the non-dimensional speed."""
        return self.flow_speed / math.sqrt(self.gravity * self.reference_length)

    @froude_num.setter
    def froude_num(self, value: float) -> None:
        self.flow_speed = value * math.sqrt(self.gravity * self.reference_length)

    @property
    def stagnation_pressure(self) -> float:
        """float: The pressure at the stagnation point."""
        return 0.5 * self.density * self.flow_speed**2

    @property
    def k0(self) -> float:
        """float: A wave number used internally in the potential-flow solver."""
        return self.gravity / self.flow_speed**2

    @property
    def lam(self) -> float:
        """float: A wavelength used internally in the potential-flow solver."""
        return 2 * math.pi / self.k0


class BodyConfig(SubConfig):
    """Configuration for the rigid body.

    Attributes:
        xCofG (float): x-coordinate of the center of gravity.
        yCofG (float): y-coordinate of the center of gravity.
        mass (float): Mass of the rigid body.
        reference_length (float): Reference length of the body.
        time_step (float): Time step to use when solving time-domain rigid body motion.
        relax_rigid_body (float): Under-relaxation factor for static rigid body motion solver.
        motion_method (str): Motion method to use for rigid body solver.
        motion_jacobian_first_step (float): Step length for first step in Jacobian calculation.
        bow_seal_tip_load (float): Fixed load to apply to bow seal tip.
        tip_constraint_ht (float): Height of constraint for seal tip.
        seal_load_pct (float): Contribution of seals as percentage of total weight.
        cushion_force_method (str): Method to use for cushion force calculation.
        initial_draft (float): Initial draft to use in motion solver.
        initial_trim (float): Initial trim to use in motion solver.
        max_draft_step (float): Maximum step to use when solving for draft.
        max_trim_step (float): Maximum step to use when solving for trim.
        max_draft_acc (float): Maximum step in acceleration to use when solving for draft.
        max_trim_acc (float): Maximum step in acceleration to use when solving for trim.
        free_in_draft (bool): If True, body will be free in draft.
        free_in_trim (bool): If True, body will be free in trim.
        draft_damping (float): Damping value to use when solving for draft.
        trim_damping (float): Damping value to use when solving for trim.

    """

    xCofG = ConfigItem(default=0.0)
    yCofG = ConfigItem(default=0.0)
    _xCofR = ConfigItem(type=float)
    _yCofR = ConfigItem(type=float)

    mass = ConfigItem("m", default=1.0)
    _weight = ConfigItem("W")

    reference_length = ConfigItem("Lref", "Lc", default=1.0)

    _cushion_pressure = ConfigItem("Pc", default=0.0)
    _seal_pressure = ConfigItem("Ps", default=0.0)

    # Rigid body motion parameters
    time_step = ConfigItem("timeStep", default=1e-3)
    relax_rigid_body = ConfigItem("rigidBodyRelax", default=1.0)
    motion_method = ConfigItem("motionMethod", default="Physical")
    motion_jacobian_first_step = ConfigItem("motionJacobianFirstStep", default=1e-6)

    bow_seal_tip_load = ConfigItem("bowSealTipLoad", default=0.0)
    tip_constraint_ht = ConfigItem("tipConstraintHt", type=float)

    seal_load_pct = ConfigItem("sealLoadPct", default=1.0)
    cushion_force_method = ConfigItem("cushionForceMethod", default="Fixed")

    initial_draft = ConfigItem("initialDraft", default=0.0)
    initial_trim = ConfigItem("initialTrim", default=0.0)

    max_draft_step = ConfigItem("maxDraftStep", default=1e-3)
    max_trim_step = ConfigItem("maxTrimStep", default=1e-3)

    max_draft_acc = ConfigItem("maxDraftAcc", default=1000.0)
    max_trim_acc = ConfigItem("maxTrimAcc", default=1000.0)

    free_in_draft = ConfigItem("freeInDraft", default=False)
    free_in_trim = ConfigItem("freeInTrim", default=False)

    draft_damping = ConfigItem("draftDamping", default=1000.0)
    trim_damping = ConfigItem("trimDamping", default=500.0)

    _relax_draft = ConfigItem("draftRelax", type=float)
    _relax_trim = ConfigItem("trimRelax", type=float)

    @property
    def relax_draft(self) -> float:
        """float: Under-relaxation factor to apply to the draft solver.

        Defaults to rigid body under-relaxation factor.

        """
        try:
            return self._relax_draft
        except AttributeError:
            return self.parent.body.relax_rigid_body

    @property
    def relax_trim(self) -> float:
        """float: Under-relaxation factor to apply to the trim solver.

        Defaults to rigid body under-relaxation factor.

        """
        try:
            return self._relax_trim
        except AttributeError:
            return self.parent.body.relax_rigid_body

    @property
    def Pc(self) -> float:
        """float: Alias for cushion pressure."""
        return self._cushion_pressure

    @property
    def PcBar(self) -> float:
        """float: Non-dimensional cushion pressure."""
        return self._cushion_pressure * self.reference_length / self.weight

    @PcBar.setter
    def PcBar(self, value: float) -> None:
        self._cushion_pressure = value * self.weight / self.reference_length

    @property
    def Ps(self) -> float:
        """float: Pressure inside the seal."""
        return self._seal_pressure

    @property
    def PsBar(self) -> float:
        """float: Non-dimensional seal pressure as ratio of cushion pressure."""
        if self._cushion_pressure == 0.0:
            return 0.0
        return self._seal_pressure / self._cushion_pressure

    @PsBar.setter
    def PsBar(self, value: float) -> None:
        self._seal_pressure = value * self._cushion_pressure

    @property
    def xCofR(self) -> float:
        """float: x-coordinate of the center of rotation. Defaults to center of gravity."""
        try:
            return self._xCofR
        except AttributeError:
            return self.xCofG

    @property
    def yCofR(self) -> float:
        """float: y-coordinate of the center of rotation. Defaults to center of gravity."""
        try:
            return self._yCofR
        except AttributeError:
            return self.yCofG

    @property
    def weight(self) -> float:
        """float: Weight of the body. Defaults to mass times gravity."""
        try:
            return self._weight
        except AttributeError:
            return self.mass * self.parent.flow.gravity

    @weight.setter
    def weight(self, value: float) -> None:
        self.mass = value / self.parent.flow.gravity


class PlotConfig(SubConfig):
    """Configuration for plotting of the response.

    Attributes:
        pType (str): Selected method by which to scale pressure lines.
        growth_rate (float): Rate at which to grow points when plotting free surface.
        CofR_grid_len (float): Length of grid for plotting center-of-rotation.
        fig_format (str): Format to save figures in.
        pressure_limiter (bool): If True, limit the pressure bar length.
        save (bool): If True, figures will be saved.
        show (bool): If True, show the plot window while running the solver.
        show_pressure (bool): If True, show the pressure profile lines.

    """

    pressure_scale_method = ConfigItem("pScaleType", default="stagnation")
    _pressure_scale = ConfigItem("pScale", default=1.0)
    _pressure_scale_pct = ConfigItem("pScalePct", default=1.0)
    _pressure_scale_head = ConfigItem("pScaleHead", default=1.0)
    growth_rate = ConfigItem("growthRate", default=1.1)
    CofR_grid_len = ConfigItem("CofRGridLen", default=0.5)
    fig_format = ConfigItem("figFormat", default="png")

    pressure_limiter = ConfigItem("pressureLimiter", default=False)

    # Load plot extents
    ext_e = ConfigItem("extE", default=0.1)
    ext_w = ConfigItem("extW", default=0.1)
    ext_n = ConfigItem("extN", default=0.1)
    ext_s = ConfigItem("extS", default=0.1)

    xmin = ConfigItem("plotXMin", type=float, default=None)
    xmax = ConfigItem("plotXMax", type=float, default=None)
    ymin = ConfigItem("plotYMin", type=float, default=None)
    ymax = ConfigItem("plotYMax", type=float, default=None)

    lambda_min = ConfigItem("lamMin", default=-1.0)
    lambda_max = ConfigItem("lamMax", default=1.0)

    _x_fs_min = ConfigItem("xFSMin", type=float, default=None)
    _x_fs_max = ConfigItem("xFSMax", type=float, default=None)

    # Whether to save, show, or watch plots
    save = ConfigItem("plotSave", default=False)
    show_pressure = ConfigItem("plotPressure", default=False)
    show = ConfigItem("plotShow", default=False)
    _watch = ConfigItem("plotWatch", default=False)

    @property
    def watch(self) -> bool:
        """bool: If True, watch the plot figure."""
        return self._watch or self.show

    @watch.setter
    def watch(self, value: bool) -> None:
        self._watch = value

    @property
    def plot_any(self) -> bool:
        """bool: If True, plot will be generated, otherwise skip to save compute time."""
        return self.show or self.save or self.watch or self.show_pressure

    @plot_any.setter
    def plot_any(self, value: bool) -> None:
        if not value:
            self.save = False
            self.show = False
            self.show_pressure = False
            self._watch = False
        else:
            raise ValueError("config.plotting.plot_any cannot be set to True")

    @property
    def x_fs_min(self) -> float:
        """float: Minimum x-location to use for plotting free surface."""
        if self._x_fs_min is not None:
            return self._x_fs_min
        if self.xmin is not None:
            return self.xmin
        return self.lambda_min * self.parent.flow.lam

    @property
    def x_fs_max(self) -> float:
        """float: Maximum x-location to use for plotting free surface."""
        if self._x_fs_max is not None:
            return self._x_fs_max
        if self.xmax is not None:
            return self.xmax
        return self.lambda_max * self.parent.flow.lam

    @property
    def pressure_scale(self) -> float:
        """float: Pressure value to use to scale the pressure profile."""
        if self.pressure_scale_method == "stagnation":
            ref_pressure = self.parent.flow.stagnation_pressure
        elif self.pressure_scale_method == "cushion":
            ref_pressure = self.parent.body.Pc if self.parent.body.Pc > 0.0 else 1.0
        elif self.pressure_scale_method == "hydrostatic":
            ref_pressure = (
                self.parent.flow.density * self.parent.flow.gravity * self._pressure_scale_head
            )
        else:
            ref_pressure = self._pressure_scale
        return ref_pressure * self._pressure_scale_pct


class PathConfig(SubConfig):
    """Directories and file formats.

    Attributes:
        case_dir (str): Path to the case directory.
        fig_dir_name (str): Path to the figures directory, relative to the case directory.
        body_dict_dir (str): Path to the body dict directory, relative to the case directory.
        input_dict_dir (str): Path to the input dict directory, relative to the case directory.
        cushion_dict_dir (str): Path to the cushion dict directory, relative to the case directory.
        mesh_dir (str): Path to the mesh directory, relative to the case directory.
        mesh_dict_dir (str): Path to the mesh dictionary file, relative to the case directory.

    Todo:
        * Verify these attribute descriptions.

    """

    case_dir = ConfigItem("caseDir", default=".")
    fig_dir_name = ConfigItem("figDirName", default="figures")
    body_dict_dir_name = ConfigItem("bodyDictDir", default="bodyDict")
    input_dict_dir_name = ConfigItem("inputDictDir", default="inputDict")
    cushion_dict_dir_name = ConfigItem(
        "pressureCushionDictDir", "cushionDictDir", default="cushionDict"
    )
    mesh_dir_name = ConfigItem("meshDir", default="mesh")
    mesh_dict_name = ConfigItem("meshDictDir", default="meshDict")


class IOConfig(SubConfig):
    """Configuration for file I/O.

    Attributes:
        data_format (str): Format of text files to save data in.
        write_interval (int): Interval in iterations for which to write result files.
        write_time_histories (bool): If True, time histories of motion will be written to files.
        results_from_file (bool): If True, load the results from previously-saved files.

    """

    data_format = ConfigItem("dataFormat", default="txt")
    write_interval = ConfigItem("writeInterval", default=1)
    write_time_histories = ConfigItem("writeTimeHistories", default=False)
    results_from_file = ConfigItem("resultsFromFile", default=False)


class SolverConfig(SubConfig):
    """Parameters for solvers.

    Attributes:
        wetted_length_solver (str): Chosen solver to use for wetted length.
        wetted_length_tol (float): Tolerance to use for wetted-length solver.
        wetted_length_relax (float): Under-relaxation factor to use for wetted-length solver.
        wetted_length_max_it (int): Maximum number of iterations to use for wetted-length solver.
        wetted_length_max_it_0 (int): Maximum number of iterations for wetted-length solver in first
            rigid body iteration.
        wetted_length_max_step_pct (float): Maximum allowable change in wetted length as fraction of
            wetted length.


    """

    wetted_length_solver = ConfigItem("wettedLengthSolver", default="Secant")
    wetted_length_tol = ConfigItem("wettedLengthTol", default=1e-6)
    wetted_length_relax = ConfigItem("wettedLengthRelax", default=1.0)
    wetted_length_max_it = ConfigItem("wettedLengthMaxIt", default=20)
    wetted_length_max_it_0 = ConfigItem("wettedLengthMaxIt0", default=100)
    wetted_length_max_step_pct = ConfigItem("wettedLengthMaxStepPct", default=0.2)
    _wetted_length_max_step_pct_inc = ConfigItem("wettedLengthMaxStepPctInc", type=float)
    _wetted_length_max_step_pct_dec = ConfigItem("wettedLengthMaxStepPctDec", type=float)
    wetted_length_max_jacobian_reset_step = ConfigItem(
        "wettedLengthMaxJacobianResetStep", default=100
    )

    max_it = ConfigItem("maxIt", default=1)
    num_ramp_it = ConfigItem("rampIt", default=0)
    relax_initial = ConfigItem("relaxI", default=0.01)
    relax_final = ConfigItem("relaxF", default=0.5)
    max_residual = ConfigItem("tolerance", default=1e-6)
    pretension = ConfigItem("pretension", default=0.1)
    relax_FEM = ConfigItem("FEMRelax", "relaxFEM", default=1.0)
    max_FEM_disp = ConfigItem("maxFEMDisp", default=1.0)
    num_damp = ConfigItem("numDamp", default=0.0)

    @property
    def wetted_length_max_step_pct_inc(self) -> float:
        """float: Maximum allowable increase in wetted length as fraction of wetted length."""
        try:
            return self._wetted_length_max_step_pct_inc
        except AttributeError:
            return self.wetted_length_max_step_pct

    @property
    def wetted_length_max_step_pct_dec(self) -> float:
        """float: Maximum allowable decrease in wetted length as fraction of wetted length."""
        try:
            return self._wetted_length_max_step_pct_dec
        except AttributeError:
            return self.wetted_length_max_step_pct


class Config:
    def __init__(self) -> None:
        self.flow = FlowConfig(parent=self)
        self.body = BodyConfig(parent=self)
        self.plotting = PlotConfig(parent=self)
        self.path = PathConfig(parent=self)
        self.io = IOConfig(parent=self)
        self.solver = SolverConfig(parent=self)

    @classmethod
    def from_file(cls, filename: Path | str) -> "Config":
        obj = cls()
        obj.load_from_file(filename)
        return obj

    def load_from_file(self, filename: Path | str) -> None:
        """Load the configuration from a file.

        Args:
            filename: The name of the file.

        """
        logger.info(f"Loading values from {filename}")
        for c in [self.flow, self.body, self.plotting, self.path, self.io, self.solver]:
            c.load_from_file(filename)
